/*

     gnuboy 1.0.3

*/



#ifndef _NES_ONLY



#include <windows.h>
#include <windowsx.h>
#include <shlwapi.h>

#include "gnuboy.h"
#include "arc.h"

#include "GB_defs.h"
#include "GB_regs.h"
#include "GB_hw.h"
#include "GB_cpu.h"
#include "GB_mem.h"
#include "GB_rc.h"
#include "GB_rtc.h"
#include "GB_reflesh.h"

#include "regdef.h"

#define GB_SOUND_GNUBOY


extern int MystrCmp(char *,char *);
extern struct gb_cpu gb_cpu;
extern void MyCtRestore();


const int crsmask[4] = { 0x7BDE, 0x739C, 0x6318, 0x4210 };

enum plstatus
{
	pl_unused = 0,
	pl_linger,
	pl_active,
	pl_locked
};


GAMEBOY::GAMEBOY(const char* ROM_name, GB_screen_mgr* _screen_mgr, sound_mgr* _sound_mgr, HWND parent_window_handle){
	scr_mgr = _screen_mgr;
	snd_mgr = _sound_mgr;
	main_window_handle = parent_window_handle;
	//	scr_mgr->setParentNES(this);
	CartRAM	=NULL;
	ROMimage=NULL;
	CartRAMSize =0;
	ROMTitle[0] =0;
	scale = 1;
	density = 0;
	rgb332=0;
	sprsort = 1;
	sprdebug=0;
//	mbc_enableram = 0;

	memset(&gb_cpu, 0, sizeof(gb_cpu));
	memset(&hw, 0, sizeof(hw));
	memset(&rtc, 0, sizeof(rtc));
	memset(&fb, 0, sizeof(fb));
	memset(&snd, 0, sizeof(snd));
	memset(&pcm, 0, sizeof(pcm));
	{
		pixmap p;
		fb.pelsize = 2;
		fb.indexed = fb.pelsize == 1;
		if(scr_mgr->lock(p)){
			fb.ptr	=	p.data;
			fb.w	=	p.width;
			fb.h	=	p.height;
			fb.pitch=	p.pitch;
			fb.enabled=	1;
			scr_mgr->unlock();
		}

		int red_length, green_length, blue_length;
		int red_offset, green_offset, blue_offset;
		red_length = 5;
		green_length = 5;
		blue_length = 5;
		red_offset = 10;
		green_offset = 5;
		blue_offset = 0;

		fb.cc[0].r = 8 - red_length;
		fb.cc[1].r = 8 - green_length;
		fb.cc[2].r = 8 - blue_length;
		fb.cc[0].l = red_offset;
		fb.cc[1].l = green_offset;
		fb.cc[2].l = blue_offset;
	}

	try {
		if(!LoadROM(main_window_handle, (char *)ROM_name))
			throw NULL;
	} catch(...) {
		if(CartRAM!=NULL){
			free(CartRAM);
			CartRAM=NULL;
		}
		if(ROMimage!=NULL){
			free(ROMimage);
			ROMimage=NULL;
		}
		throw;
	}
//	reset;
//	gameboyp = this;
	reset(0);
}





//------------------------------



BOOL GAMEBOY::LoadROM(HWND hWnd, char *FileName)
{
	HANDLE hFile;
	DWORD dwBytes ,FileSize;
	unsigned char *pt = NULL;
	const char *GB_Extentions[] = { "*.gb", "*.gbc",NULL};

	hFile = CreateFile(FileName, GENERIC_READ, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if(hFile == INVALID_HANDLE_VALUE)
		return 0;
	pt = (unsigned char *)malloc(0x200);
	if(pt==NULL){
		CloseHandle(hFile);
		return 0;
	}
	ReadFile(hFile, pt, 0x200, &dwBytes, NULL);

	if(*(DWORD *)(&pt[0x104]) != 0x6666EDCE){
		CloseHandle(hFile);
		free(pt);
		if(!Uncompress(NULL, &pt, FileName, (LPDWORD)&FileSize, GB_Extentions))
			return 0;
		if(*(DWORD *)&pt[0x104] != 0x6666EDCE){
			free(pt);
			return 0;
		}
	}
	else{
		unsigned char *t = pt;
		FileSize = GetFileSize(hFile, NULL);
		pt = (unsigned char *)realloc(t, FileSize);
		if(pt==NULL){
			CloseHandle(hFile);
			return 0;
		}
		ReadFile(hFile, pt+0x200, FileSize-0x200, &dwBytes, NULL);
		CloseHandle(hFile);
	}

	memcpy(ROMTitle, &pt[0x134], 0x10);
	if(ROMTitle[0x0e] & 0x80) ROMTitle[14] = 0;
	if(ROMTitle[0x0f] & 0x80) ROMTitle[15] = 0;
	ROMTitle[0x10] = 0;

	{
		int ctype = pt[0x147];
		rtc.batt = 0;
		switch(ctype){
			case 1: case 2: case 3:
				mbc_type = 1;
				break;
			case 5: case 6:
				mbc_type = 2;
				break;
			case 0x0f: case 0x10:
				rtc.batt = 1;
			case 0x11: case 0x12: case 0x13:
				mbc_type = 3;
				break;
			case 0x19: case 0x1a: case 0x1b:
				mbc_type = 3;
				break;
			case 0x1c: case 0x1d: case 0x1e:
				mbc_type = MBC_RUMBLE;
				break;
			case 0x0fe:
				mbc_type = MBC_HUC3;
				break;
			case 0x0ff:
				mbc_type = MBC_HUC1;
				break;
			default:
				mbc_type = 0;
//				free(pt);
//				return 0;
				break;
		}
		switch(ctype){
			case 3: case 6: case 9: case 0x0d: case 0x10: case 0x12: case 0x13: case 0x1b: case 0x1e:
				mbc_batt = 1;
//				mbc.batt = (batt_table[c] && !nobatt) || forcebatt;
				break;
			default:
				mbc_batt = 0;
				break;
		}
		ctype = pt[0x148];
		switch(ctype){
			case 0: mbc_romsize = 2;
				break;
			case 1: mbc_romsize = 4;
				break;
			case 2: mbc_romsize = 8;
				break;
			case 3: mbc_romsize = 16;
				break;
			case 4: mbc_romsize = 32;
				break;
			case 5: mbc_romsize = 64;
				break;
			case 6: mbc_romsize = 128;
				break;
			case 7: mbc_romsize = 256;
				break;
			case 8: mbc_romsize = 512;
				break;
			case 0x53: case 0x54: case 0x55:
				mbc_romsize = 128;
				break;
			default:
				free(pt);
				return 0;
				break;

		}

		ctype = pt[0x149];
		switch(ctype){
			case 0: case 1: case 2:
				mbc_ramsize = 1;
				break;
			case 3: case 5:
				mbc_ramsize = 4;
				break;
			case 4:
				mbc_ramsize = 1;
				break;
			default:
				free(pt);
				return 0;
				break;
		}
	}

//	rtc.batt = rtc_table[c];
/*	mbc_romsize = romsize_table[header[0x0148]];
	mbc_ramsize = ramsize_table[header[0x0149]];

	if (!mbc_romsize) die("unknown ROM size %02X\n", header[0x0148]);
	if (!mbc_ramsize) die("unknown SRAM size %02X\n", header[0x0149]);

	rlen = 16384 * mbc_romsize;
	rom.bank = realloc(data, rlen);
	if (rlen > len) memset(rom.bank[0]+len, 0xff, rlen - len);
*/	
//	ram.sbank = (char *)malloc(8192 * mbc_ramsize);

	CartRAMSize = 8192 * mbc_ramsize;
	CartRAM =(unsigned char *)malloc(CartRAMSize);
	memset(CartRAM, 0, CartRAMSize);
	ROMimage=pt;

//	initmem(ram.sbank, 8192 * mbc_ramsize);
//	initmem(ram.ibank, 4096 * 8);


	hw.cgb = ((pt[0x143] == 0x80) || (pt[0x143] == 0xc0))/* && !forcedmg*/;
	hw.gba = 0; //(hw.cgb /*&& gbamode*/);

	strcpy(ROMPath, FileName);
	PathRemoveFileSpec(ROMPath);
	{
		char *fnp = PathFindFileName(FileName);
		if(fnp!=NULL){
			strcpy(ROMname, fnp);
			PathRemoveExtension(ROMname);
		}
		else{
			strcpy(ROMname, "n/a");
		}
	}
	Load_SaveRAM();
	return TRUE;
}



GAMEBOY::~GAMEBOY(){
	Save_SaveRAM();
	if(CartRAM!=NULL){
		free(CartRAM);
		CartRAM=NULL;
	}
	if(ROMimage!=NULL){
		free(ROMimage);
		ROMimage=NULL;
	}
//	if(pcm.buf)
//		free(pcm.buf);

	return;
}

//extern void gb_cpu_reset(GAMEBOY *gb);
extern int gb_cpu_emulate(int cycles);


void GAMEBOY::reset(unsigned char f)
{
	cpu_reset();
	hw_reset();
	lcd_reset();
	mbc_reset();
	sound_reset();

	memset(MainMem, 0, 0x4000);
	memset(Hi_Memory, 0, 0x200);

#if 0
	mem_write(0xFF05, 00);
	mem_write(0xFF06, 00);
	mem_write(0xFF07, 00);

	mem_write(0xFF10, 0x80);
	mem_write(0xFF11, 0xBF);
	mem_write(0xFF12, 0xF3);
	mem_write(0xFF14, 0xBF);

#endif
	mem_write(0xFF40, 0x91);
	mem_write(0xFF42, 0x00);
	mem_write(0xFF43, 0x00);
	mem_write(0xFF45, 0x00);
	mem_write(0xFF47, 0xFC);
	mem_write(0xFF48, 0xFF);
	mem_write(0xFF49, 0xFF);
	mem_write(0xFF4A, 0x00);
	mem_write(0xFF4B, 0x00);

	mem_write(0xFFFF, 0x00);
}

void GAMEBOY::cpu_reset()
{
	memset(&gb_cpu, 0, sizeof(gb_cpu));

	gb_cpu.gameboy = this;
	gb_cpu.himem = Hi_Memory;

	gb_cpu.speed = 0;
	gb_cpu.halt = 0;
	gb_cpu.div = 0;
	gb_cpu.tim = 0;
	gb_cpu.lcdc = 40;

	gb_cpu.ima = 0;
	gb_cpu.ime = 0;

	gb_cpu.pc.w[0] = 0x0100;
	gb_cpu.sp.w[0] = 0xFFFE;
	gb_cpu.af.w[0] = 0x01B0;
	gb_cpu.bc.w[0] = 0x0013;
	gb_cpu.de.w[0] = 0x00D8;
	gb_cpu.hl.w[0] = 0x014D;

#ifdef _DEBUG
//	debug_trace = 1;
#endif
	if(hw.cgb)
		gb_cpu.af.b[0][1] = 0x11;
//	if(hw.gba)
//		B = 0x01;

	mbc_rombank = 1;
	mbc_rambank = 0;

	gb_cpu.mem_page[0x00] = ROMimage;
	gb_cpu.mem_page[0x01] = ROMimage + 0x1000;
	gb_cpu.mem_page[0x02] = ROMimage + 0x2000;
	gb_cpu.mem_page[0x03] = ROMimage + 0x3000;
	gb_cpu.mem_page[0x04] = ROMimage + 0x4000;
	gb_cpu.mem_page[0x05] = ROMimage + 0x5000;
	gb_cpu.mem_page[0x06] = ROMimage + 0x6000;
	gb_cpu.mem_page[0x07] = ROMimage + 0x7000;
	gb_cpu.mem_page[0x08] = NULL;
	gb_cpu.mem_page[0x09] = NULL;
	gb_cpu.mem_page[0x0a] = CartRAM;
	gb_cpu.mem_page[0x0b] = CartRAM+0x1000;
	gb_cpu.mem_page[0x0c] = MainMem;
	gb_cpu.mem_page[0x0d] = MainMem+0x1000;
	gb_cpu.mem_page[0x0e] = NULL;
	gb_cpu.mem_page[0x0f] = NULL;

}


void GAMEBOY::Save_SaveRAM(){
	char str[MAX_PATH];
	HANDLE hFile;
	DWORD dwAccBytes, dw;

	if(CartRAM==NULL || mbc_batt == 0)
		return;
	GetModuleFileName(NULL, str, MAX_PATH);
	PathRemoveFileSpec(str);
	PathCombine(str, str, "GBsave");

	dw = GetFileAttributes(str);
	if(dw == 0xFFFFFFFF){
		dw = GetLastError();
		if(!(dw & FILE_ATTRIBUTE_DIRECTORY)){
			if(!CreateDirectory(str, NULL))
				return;
		}
	}
	
	PathCombine(str , str, ROMname);
	strcat(str, ".sav");
	hFile = CreateFile(str, GENERIC_WRITE, 0, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if(hFile == INVALID_HANDLE_VALUE){
		return;
    }
	WriteFile(hFile, CartRAM, CartRAMSize, &dwAccBytes, NULL);
	CloseHandle(hFile);
}


void GAMEBOY::Load_SaveRAM(){
	char str[MAX_PATH];
	HANDLE hFile;
	DWORD dwAccBytes;

	if(CartRAM==NULL)
		return;
	GetModuleFileName(NULL, str, MAX_PATH);
	PathRemoveFileSpec(str);
	PathCombine(str, str, "GBsave");

	PathCombine(str , str, ROMname);
	strcat(str, ".sav");
	hFile = CreateFile(str, GENERIC_READ, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if(hFile == INVALID_HANDLE_VALUE){
		return;
    }
	ReadFile(hFile, CartRAM, CartRAMSize, &dwAccBytes, NULL);
	CloseHandle(hFile);
}


boolean GAMEBOY::loadState(const char* fn){
	return 0;
}


// ID 4byte, size 4byte
//
// REGS 0x53474552		
//		PC, SP, BC, DE, HL, AF, IME, ima, spd
// RTC  0x20435452		
//
// SND	0x00444E53
//
// HRAM	0x	
//

boolean GAMEBOY::saveState(const char* fn){
	char str[MAX_PATH], *pfn;
	HANDLE hFile;
	DWORD dwAccBytes;

	return 0;
#if 0
	struct svar svars[] = 
	{
		I4("GbSs", &ver),
			
			I2("PC  ", &PC),
			I2("SP  ", &SP),
			I2("BC  ", &BC),
			I2("DE  ", &DE),
			I2("HL  ", &HL),
			I2("AF  ", &AF),
			
			I4("IME ", &cpu.ime),
			I4("ima ", &cpu.ima),
			I4("spd ", &cpu.speed),
			I4("halt", &cpu.halt),
			I4("div ", &cpu.div),
			I4("tim ", &cpu.tim),
			I4("lcdc", &cpu.lcdc),
			I4("snd ", &cpu.snd),
			
			I1("ints", &hw.ilines),
			I1("pad ", &hw.pad),
			I4("cgb ", &hw.cgb),
			I4("gba ", &hw.gba),
			
			I4("mbcm", &mbc.model),
			I4("romb", &mbc.rombank),
			I4("ramb", &mbc.rambank),
			I4("enab", &mbc.enableram),
			I4("batt", &mbc.batt),
			
			I4("rtcR", &rtc.sel),
			I4("rtcL", &rtc.latch),
			I4("rtcC", &rtc.carry),
			I4("rtcS", &rtc.stop),
			I4("rtcd", &rtc.d),
			I4("rtch", &rtc.h),
			I4("rtcm", &rtc.m),
			I4("rtcs", &rtc.s),
			I4("rtct", &rtc.t),
			I1("rtR8", &rtc.regs[0]),
			I1("rtR9", &rtc.regs[1]),
			I1("rtRA", &rtc.regs[2]),
			I1("rtRB", &rtc.regs[3]),
			I1("rtRC", &rtc.regs[4]),
			
			I4("S1on", &snd.ch[0].on),
			I4("S1p ", &snd.ch[0].pos),
			I4("S1c ", &snd.ch[0].cnt),
			I4("S1ec", &snd.ch[0].encnt),
			I4("S1sc", &snd.ch[0].swcnt),
			I4("S1sf", &snd.ch[0].swfreq),
			
			I4("S2on", &snd.ch[1].on),
			I4("S2p ", &snd.ch[1].pos),
			I4("S2c ", &snd.ch[1].cnt),
			I4("S2ec", &snd.ch[1].encnt),
			
			I4("S3on", &snd.ch[2].on),
			I4("S3p ", &snd.ch[2].pos),
			I4("S3c ", &snd.ch[2].cnt),
			
			I4("S4on", &snd.ch[3].on),
			I4("S4p ", &snd.ch[3].pos),
			I4("S4c ", &snd.ch[3].cnt),
			I4("S4ec", &snd.ch[3].encnt),
			
			I4("hdma", &hw.hdma),
			
			I4("sram", &sramblock),
			I4("iram", &iramblock),
			I4("vram", &vramblock),
			I4("hi  ", &hiofs),
			I4("pal ", &palofs),
			I4("oam ", &oamofs),
			I4("wav ", &wavofs),
			
			/* NOSAVE is a special code to prevent the rest of the table
			* from being saved, used to support old stuff for backwards
			* compatibility... */
			NOSAVE,
			
			/* the following are obsolete as of 0x104 */
			
			I4("hram", &hramofs),
			
			R(P1), R(SB), R(SC),
			R(DIV), R(TIMA), R(TMA), R(TAC),
			R(IE), R(IF),
			R(LCDC), R(STAT), R(LY), R(LYC),
			R(SCX), R(SCY), R(WX), R(WY),
			R(BGP), R(OBP0), R(OBP1),
			R(DMA),
			
			R(VBK), R(SVBK), R(KEY1),
			R(BCPS), R(BCPD), R(OCPS), R(OCPD),
			
			R(NR10), R(NR11), R(NR12), R(NR13), R(NR14),
			R(NR21), R(NR22), R(NR23), R(NR24),
			R(NR30), R(NR31), R(NR32), R(NR33), R(NR34),
			R(NR41), R(NR42), R(NR43), R(NR44),
			R(NR50), R(NR51), R(NR52),
			
			I1("DMA1", &R_HDMA1),
			I1("DMA2", &R_HDMA2),
			I1("DMA3", &R_HDMA3),
			I1("DMA4", &R_HDMA4),
			I1("DMA5", &R_HDMA5),
			
			END
	};

	if(CartRAM==NULL)
		return 0;
	pfn = PathFindFileName(fn);
	GetModuleFileName(NULL, str, MAX_PATH);
	PathRemoveFileSpec(str);
	PathCombine(str, str, "GBssave");

	dw = GetFileAttributes(str);
	if(!(dw & FILE_ATTRIBUTE_DIRECTORY)){	//tH_
		if(!CreateDirectory(str, NULL))		//create
			return 0;
	}

	PathCombine(str , str, fn);
	hFile = CreateFile(str, GENERIC_WRITE, 0, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if(hFile == INVALID_HANDLE_VALUE){
		return;
    }
	WriteFile(hFile, CartRAM, CartRAMSize, &dwAccBytes, NULL);
	CloseHandle(hFile);
#endif
	return 0;
}

/*
static int framelen = 16743;
static int framecount;
*/
extern int gb_sndclock;


boolean GAMEBOY::emulate_frame(boolean draw)
{
	int lockf = 0;
//---	vid_begin();
	MyCtRestore();

	pixmap p;
	fb.dirty = 0;
	if(draw && (lockf = scr_mgr->lock(p))){
		fb.ptr	=	p.data;
		fb.w	=	p.width;
		fb.h	=	p.height;
		fb.pitch=	p.pitch;
		/// 16bit
//		int red_length, green_length, blue_length;
//		int red_offset, green_offset, blue_offset;
/*
		red_length = 8;
		green_length = 8;
		blue_length = 8;
		red_offset = 16;
		green_offset = 8;
		blue_offset = 0;
*/
		fb.enabled=	1;
	}
	else{
		fb.enabled=	0;
	}
	lcd_begin();
	pad_check();
	gb_cpu_emulate(2280);
	while(R_LY > 0 && R_LY < 144){
		gb_cpu_emulate(gb_cpu.lcdc);
//		int ly = R_LY;		/// test
	}
	{
			if(lockf){
				scr_mgr->unlock();
				lockf = 0;
				fb.enabled=	0;
			}
			rtc_tick();
			if (!(R_LCDC & 0x80))
				gb_cpu_emulate(32832);
			while(R_LY > 0) /* wait for next frame */
				gb_cpu_emulate(gb_cpu.lcdc);
			apu_process();
#ifdef _DEBUG
//			LOG("frame sound clock  " << HEX(gb_sndclock,8) <<endl);
			gb_sndclock = 0;
#endif
	}
	return 0;
}


/*
struct mbc mbc;

struct rom rom;
struct ram ram;
*/

/*
 * In order to make reads and writes efficient, we keep tables
 * (indexed by the high nibble of the address) specifying which
 * regions can be read/written without a function call. For such
 * ranges, the pointer in the map table points to the base of the
 * region in host system memory. For ranges that require special
 * processing, the pointer is NULL.
 *
 * mem_updatemap is called whenever bank changes or other operations
 * make the old maps potentially invalid.
 */

#define SetROMBank4000H(bn){\
	n = bn; \
	if(!n) \
		n=1; \
	n*=0x4000; \
	gb_cpu.mem_page[0x04] = ROMimage + n; \
	n+=0x1000; \
	gb_cpu.mem_page[0x05] = ROMimage + n; \
	n+=0x1000; \
	gb_cpu.mem_page[0x06] = ROMimage + n; \
	n+=0x1000; \
	gb_cpu.mem_page[0x07] = ROMimage + n; }

#define SetRAMBankA000H(bn) { n=(bn)*0x2000; \
	gb_cpu.mem_page[0x0a] = CartRAM + n; \
	gb_cpu.mem_page[0x0b] = CartRAM + n +0x1000; }




void GAMEBOY::mem_updatemap()
{
	int n;

	SetROMBank4000H(mbc_rombank);
#ifdef _DEBUG
//	LOG("0x4000 " << HEX(mbc_rombank,8) << endl);
//	LOG("0x4000 " << HEX(n,8) << endl);
#endif
	SetRAMBankA000H(mbc_rambank);
#ifdef _DEBUG
//	LOG("0xA000 " << HEX(mbc_rambank,8) << endl);
//	LOG("0xA000 " << HEX(n,8) << endl);
#endif

/*
	switch (mbc_type)
	{
	case MBC_NONE:
		break;
	case MBC_MBC1:
		if(mbc_model){
			// ram
			SetRAMBankA000H(mbc_rambank & 3);
			// rom
			SetROMBank4000H(mbc_rombank&0x1f);
		}
		else{
			// ram
			gb_cpu.mem_page[0x0a] = MainMem;
			gb_cpu.mem_page[0x0b] = MainMem+0x1000;
			// rom
			SetROMBank4000H(mbc_rombank&0x7f);
		}
		break;
	case MBC_MBC2:
		SetROMBank4000H(mbc_rombank&0x1f);
		break;
	case MBC_MBC3:
		SetRAMBankA000H(mbc_rambank & 0x0f);
		SetROMBank4000H(mbc_rombank&0x1f);
		break;
	case MBC_MBC5:
		SetRAMBankA000H(mbc_rambank & 0x0f);
		SetROMBank4000H(mbc_rombank&0x1f);
		break;
	case MBC_RUMBLE:
		SetRAMBankA000H(mbc_rambank & 0x07);
		SetROMBank4000H(mbc_rombank&0x1ff);
		break;
	case MBC_HUC1:
		break;
	case MBC_HUC3:
		break;
	}
/*	map = mbc.rmap;
	map[0x0] = rom.bank[0];
	map[0x1] = rom.bank[0];
	map[0x2] = rom.bank[0];
	map[0x3] = rom.bank[0];
	if (mbc.rombank < mbc_romsize)
	{
		map[0x4] = rom.bank[mbc.rombank] - 0x4000;
		map[0x5] = rom.bank[mbc.rombank] - 0x4000;
		map[0x6] = rom.bank[mbc.rombank] - 0x4000;
		map[0x7] = rom.bank[mbc.rombank] - 0x4000;
	}
	else map[0x4] = map[0x5] = map[0x6] = map[0x7] = NULL;
	if (0 && (R_STAT & 0x03) == 0x03)
	{
		map[0x8] = NULL;
		map[0x9] = NULL;
	}
	else
	{
		map[0x8] = lcd.vbank[R_VBK & 1] - 0x8000;
		map[0x9] = lcd.vbank[R_VBK & 1] - 0x8000;
	}
	if (mbc_enableram && !(rtc.sel&8))
	{
		map[0xA] = ram.sbank[mbc.rambank] - 0xA000;
		map[0xB] = ram.sbank[mbc.rambank] - 0xA000;
	}
	else map[0xA] = map[0xB] = NULL;
	map[0xC] = ram.ibank[0] - 0xC000;
	n = R_SVBK & 0x07;
	map[0xD] = ram.ibank[n?n:1] - 0xD000;
	map[0xE] = ram.ibank[0] - 0xE000;
	map[0xF] = NULL;
	
	map = mbc.wmap;
	map[0x0] = map[0x1] = map[0x2] = map[0x3] = NULL;
	map[0x4] = map[0x5] = map[0x6] = map[0x7] = NULL;
	map[0x8] = map[0x9] = NULL;
	if (mbc_enableram && !(rtc.sel&8))
	{
		map[0xA] = ram.sbank[mbc.rambank] - 0xA000;
		map[0xB] = ram.sbank[mbc.rambank] - 0xA000;
	}
	else map[0xA] = map[0xB] = NULL;
	map[0xC] = ram.ibank[0] - 0xC000;
	n = R_SVBK & 0x07;
	map[0xD] = ram.ibank[n?n:1] - 0xD000;
	map[0xE] = ram.ibank[0] - 0xE000;
	map[0xF] = NULL;
*/
}


/*
 * ioreg_write handles output to io registers in the FF00-FF7F,FFFF
 * range. It takes the register number (low byte of the address) and a
 * byte value to be written.
 */

void GAMEBOY::ioreg_write(int r, byte b)
{
	if (!hw.cgb)
	{
		switch (r)
		{
		case RI_VBK:
		case RI_BCPS:
		case RI_OCPS:
		case RI_BCPD:
		case RI_OCPD:
		case RI_SVBK:
		case RI_KEY1:
		case RI_HDMA1:
		case RI_HDMA2:
		case RI_HDMA3:
		case RI_HDMA4:
		case RI_HDMA5:
			return;
		}
	}
	
	switch(r)
	{
	case RI_TIMA:
	case RI_TMA:
	case RI_TAC:
	case RI_SCY:
	case RI_SCX:
	case RI_WY:
	case RI_WX:
		REG(r) = b;
		break;
	case RI_BGP:
		if (R_BGP == b) break;
		pal_write_dmg(0, 0, b);
		pal_write_dmg(8, 1, b);
		R_BGP = b;
		break;
	case RI_OBP0:
		if (R_OBP0 == b) break;
		pal_write_dmg(64, 2, b);
		R_OBP0 = b;
		break;
	case RI_OBP1:
		if (R_OBP1 == b) break;
		pal_write_dmg(72, 3, b);
		R_OBP1 = b;
		break;
	case RI_IF:
	case RI_IE:
		REG(r) = b & 0x1F;
		break;
	case RI_P1:
		REG(r) = b;
		pad_refresh();
		break;
	case RI_SC:
		/* FIXME - this is a hack for stupid roms that probe serial */
		if ((b & 0x81) == 0x81)
		{
			R_SB = 0xff;
			hw_interrupt(IF_SERIAL, IF_SERIAL);
			hw_interrupt(0, IF_SERIAL);
		}
		R_SC = b; /* & 0x7f; */
		break;
	case RI_DIV:
		REG(r) = 0;
		break;
	case RI_LCDC:
		lcdc_change(b);
		break;
	case RI_STAT:
		stat_write(b);
		break;
	case RI_LYC:
		REG(r) = b;
		stat_trigger();
		break;
	case RI_VBK:	// vram bank select
		REG(r) = b | 0xFE;
//		mem_updatemap();
		break;
	case RI_BCPS:
		R_BCPS = b & 0xBF;
		R_BCPD = lcd.pal[b & 0x3F];
		break;
	case RI_OCPS:
		R_OCPS = b & 0xBF;
		R_OCPD = lcd.pal[64 + (b & 0x3F)];
		break;
	case RI_BCPD:
		R_BCPD = b;
		pal_write(R_BCPS & 0x3F, b);
		if (R_BCPS & 0x80) R_BCPS = (R_BCPS+1) & 0xBF;
		break;
	case RI_OCPD:
		R_OCPD = b;
		pal_write(64 + (R_OCPS & 0x3F), b);
		if (R_OCPS & 0x80) R_OCPS = (R_OCPS+1) & 0xBF;
		break;
	case RI_SVBK:	// ram bank select
		REG(r) = b & 0x07;
		{
			int data= b&7;
			if(data==0)
				data=1;
			gb_cpu.mem_page[0x0d] = MainMem + (0x1000 * data);
		}
//		mem_updatemap();
		break;
	case RI_DMA:
		hw_dma(b);
		break;
	case RI_KEY1:
		REG(r) = (REG(r) & 0x80) | (b & 0x01);
		break;
	case RI_HDMA1:
		REG(r) = b;
		break;
	case RI_HDMA2:
		REG(r) = b & 0xF0;
		break;
	case RI_HDMA3:
		REG(r) = b & 0x1F;
		break;
	case RI_HDMA4:
		REG(r) = b & 0xF0;
		break;
	case RI_HDMA5:
		hw_hdma_cmd(b);
		break;
	}
	switch (r)
	{
	case RI_BGP:
	case RI_OBP0:
	case RI_OBP1:
		/* printf("palette reg %02X write %02X at LY=%02X\n", r, b, R_LY); */
	case RI_HDMA1:
	case RI_HDMA2:
	case RI_HDMA3:
	case RI_HDMA4:
	case RI_HDMA5:
		/* printf("HDMA %d: %02X\n", r - RI_HDMA1 + 1, b); */
		break;
	}
	/* printf("reg %02X => %02X (%02X)\n", r, REG(r), b); */
}


byte GAMEBOY::ioreg_read(int r)
{
	switch(r)
	{
	case RI_SC:
		r = R_SC;
		R_SC &= 0x7f;
		return r;
	case RI_P1:
	case RI_SB:
	case RI_DIV:
	case RI_TIMA:
	case RI_TMA:
	case RI_TAC:
	case RI_LCDC:
	case RI_STAT:
	case RI_SCY:
	case RI_SCX:
	case RI_LY:
	case RI_LYC:
	case RI_BGP:
	case RI_OBP0:
	case RI_OBP1:
	case RI_WY:
	case RI_WX:
	case RI_IE:
	case RI_IF:
		return REG(r);
	case RI_VBK:
	case RI_BCPS:
	case RI_OCPS:
	case RI_BCPD:
	case RI_OCPD:
	case RI_SVBK:
	case RI_KEY1:
	case RI_HDMA1:
	case RI_HDMA2:
	case RI_HDMA3:
	case RI_HDMA4:
	case RI_HDMA5:
		if (hw.cgb) return REG(r);
	default:
		return REG(r);
//		return 0xff;
	}
}



/*
 * Memory bank controllers typically intercept write attempts to
 * 0000-7FFF, using the address and byte written as instructions to
 * change rom or sram banks, control special hardware, etc.
 *
 * mbc_write takes an address (which should be in the proper range)
 * and a byte value written to the address.
 */

void GAMEBOY::mbc_write(int a, byte b)
{
	byte ha = (a>>12);

#ifdef _DEBUG

//	LOG("w  adr" << HEX(a,4)<< "w " << HEX(b,2) << endl);
#endif

	/* printf("mbc %d: rom bank %02X -[%04X:%02X]-> ", mbc_type, mbc.rombank, a, b); */
	switch (mbc_type)
	{
	case MBC_MBC1:
		switch (ha & 0xE)
		{
		case 0x0:
			mbc_enableram = ((b & 0x0F) == 0x0A);
			break;
		case 0x2:
//			if((b & 0x1F) == 0)
//				b = 0x01;
			mbc_rombank = (mbc_rombank & 0x60) | (b & 0x1F);
			break;
		case 0x4:
			if(mbc_model)
			{
				mbc_rambank = b & 0x03;
				break;
			}
			mbc_rombank = (mbc_rombank & 0x1F) | ((int)(b&3)<<5);
			break;
		case 0x6:
			mbc_model = b & 0x1;
/*
			if(mbc_model){
				mbc_rambank = 0;
				mbc_rombank&=0x1f;
			}
*/
			break;
		}
		break;
	case MBC_MBC2: /* is this at all right? */
		if ((a & 0x0100) == 0x0000)
		{
			mbc_enableram = ((b & 0x0F) == 0x0A);
			break;
		}
		if ((a & 0xE100) == 0x2100)
		{
#ifdef _DEBUG
//		LOG("MBC2  " << HEX(a,4)<< "w " << HEX(b,2) << endl);
#endif
			mbc_rombank = b & 0x0F;
			break;
		}
		break;
	case MBC_MBC3:
		switch (ha & 0xE)
		{
		case 0x0:
			mbc_enableram = ((b & 0x0F) == 0x0A);
			break;
		case 0x2:
			if ((b & 0x7F) == 0) b = 0x01;
			mbc_rombank = b & 0x7F;
			break;
		case 0x4:
			rtc.sel = b & 0x0f;
			mbc_rambank = b & 0x03;
			break;
		case 0x6:
			rtc_latch(b);
			break;
		}
		break;
	case MBC_RUMBLE:
		switch (ha & 0xF)
		{
		case 0x4:
		case 0x5:
			/* FIXME - save high bit as rumble state */
			/* mask off high bit */
			b &= 0x7;
			break;
		}
		/* fall thru */
	case MBC_MBC5:
		switch (ha & 0xF)
		{
		case 0x0:
		case 0x1:
			mbc_enableram = ((b & 0x0F) == 0x0A);
			break;
		case 0x2:
			if ((b & 0xFF) == 0) b = 0x01;
			mbc_rombank = (mbc_rombank & 0x100) | (b & 0xFF);
			break;
		case 0x3:
			mbc_rombank = (mbc_rombank & 0xFF) | ((int)(b&1)<<8);
			break;
		case 0x4:
		case 0x5:
			mbc_rambank = b & 0x0f;
			break;
		}
		break;
	case MBC_HUC1: /* FIXME - this is all guesswork -- is it right??? */
		switch (ha & 0xE)
		{
		case 0x0:
			mbc_enableram = ((b & 0x0F) == 0x0A);
			break;
		case 0x2:
			if ((b & 0x1F) == 0) b = 0x01;
			mbc_rombank = (mbc_rombank & 0x60) | (b & 0x1F);
			break;
		case 0x4:
			if (mbc_model)
			{
				mbc_rambank = b & 0x03;
				break;
			}
			mbc_rombank = (mbc_rombank & 0x1F) | ((int)(b&3)<<5);
			break;
		case 0x6:
			mbc_model = b & 0x1;
			break;
		}
		break;
	case MBC_HUC3:
		switch (ha & 0xE)
		{
		case 0x0:
			mbc_enableram = ((b & 0x0F) == 0x0A);
			break;
		case 0x2:
			b &= 0x7F;
			mbc_rombank = b ? b : 1;
			break;
		case 0x4:
			rtc.sel = b & 0x0f;
			mbc_rambank = b & 0x03;
			break;
		case 0x6:
			rtc_latch(b);
			break;
		}
		break;
	}
	mbc_rombank &= (mbc_romsize - 1);
	mbc_rambank &= (mbc_ramsize - 1);
#ifdef _DEBUG
//		LOG("222  " << HEX(mbc_rombank,2) << endl);
#endif
	/* printf("%02X\n", mbc.rombank); */
	mem_updatemap();
}


/*
 * mem_write is the basic write function. Although it should only be
 * called when the write map contains a NULL for the requested address
 * region, it accepts writes to any address.
 */

void GAMEBOY::mem_write(int a, byte b)
{
//	int n;
	byte ha = (a>>12) & 0xE;
#ifdef _DEBUG
/*
	if(a >= 0x10000){
		int i= 0;	//test
	}
*/
//	LOG("w  adr" << HEX(a,4)<< "w " << HEX(b,2) << endl);
#endif


	/* printf("write to 0x%04X: 0x%02X\n", a, b); */
	switch (ha)
	{
	case 0x0:
	case 0x2:
	case 0x4:
	case 0x6:
		mbc_write(a, b);
		break;
	case 0x8:
		/* if ((R_STAT & 0x03) == 0x03) break; */
		vram_write(a & 0x1FFF, b);
		break;
	case 0xA:
		if (!mbc_enableram) break;
		if (rtc.sel&8)
		{
			rtc_write(b);
			break;
		}
		gb_cpu.mem_page[0x0a][a & 0x1FFF] = b;
		break;
	case 0xC:
		if ((a & 0xF000) == 0xC000)
		{
			gb_cpu.mem_page[0x0c][a & 0x0FFF] = b;
			break;
		}

			if((a&0x0FFF) == 0x072){
				int i= 0;	//test
			}

		gb_cpu.mem_page[0x0d][a & 0x0FFF] = b;
//		n = R_SVBK & 0x07;
//		ram.ibank[n?n:1][a & 0x0FFF] = b;
		break;
	case 0xE:
		if (a < 0xFE00)
		{
			mem_write(a & 0xDFFF, b);
			break;
		}

		if ((a & 0xFF00) == 0xFE00)
		{
			/* if (R_STAT & 0x02) break; */
			if (a < 0xFEA0) lcd.oam.mem[a & 0xFF] = b;
			break;
		}
		/* return writehi(a & 0xFF, b); */
		if (a >= 0xFF10 && a <= 0xFF3F)
		{
//			sound_write(a & 0x1FF, b);
//			Hi_Memory[a & 0x1FF] = b;
			gb_apudata_t	t;
			t.address	= a & 0x1FF;
			t.timestamp	= gb_cpu.snd;
			t.value		= b;
			apu_enqueue(&t);
			break;
		}
		if ((a & 0xFF80) == 0xFF80/* && a != 0xFFFF*/)
		{
			Hi_Memory[a & 0x1FF] = b;
			if(a == 0xFFFF)
				ioreg_write(a & 0x1FF, b);
			break;
		}
		ioreg_write(a & 0x1FF, b);
	}
}


/*
 * mem_read is the basic read function...not useful for much anymore
 * with the read map, but it's still necessary for the final messy
 * region.
 */

byte GAMEBOY::mem_read(int a)
{
//	int n;
	byte ha = (a>>12) & 0xE;
	
	// test 
//	unsigned char *p = gb_cpu.mem_page[0];

	/* printf("read %04x\n", a); */
	switch (ha)
	{
	case 0x0:
	case 0x2:
		return gb_cpu.mem_page[0][a & 0x3FFF];
	case 0x4:
	case 0x6:
		return gb_cpu.mem_page[4][a & 0x3FFF];
	case 0x8:
		/* if ((R_STAT & 0x03) == 0x03) return 0xFF; */
		return lcd.vbank[R_VBK&1][a & 0x1FFF];
	case 0xA:
/*---
		if (!mbc_enableram && mbc_type == MBC_HUC3)
			return 0x01;
		if (!mbc_enableram)
			return 0xFF;
**/
		if (rtc.sel&8)
			return rtc.regs[rtc.sel&7];
		if(!gb_cpu.mem_page[0x0a])
			return 0xFF;
		return gb_cpu.mem_page[0x0a][a & 0x1FFF];
	case 0xC:
		if((a & 0xF000) == 0xC000){
			return gb_cpu.mem_page[0x0c][a & 0x0FFF];
		}
		else{
/*
			if((a&0x0FFF) == 0x072){
				int i= 0;	//test
			}
*/
			return gb_cpu.mem_page[0x0d][a & 0x0FFF];
		}
//			return ram.ibank[0][a & 0x0FFF];
//		n = R_SVBK & 0x07;
//		return ram.ibank[n?n:1][a & 0x0FFF];
		
	case 0xE:
		if (a < 0xFE00) return mem_read(a & 0xDFFF);
		if ((a & 0xFF00) == 0xFE00)
		{
			/* if (R_STAT & 0x02) return 0xFF; */
			if (a < 0xFEA0) return lcd.oam.mem[a & 0xFF];
			return 0xFF;
		}
		/* return readhi(a & 0xFF); */
		if (a == 0xFFFF) return REG(0x1FF);
		if (a >= 0xFF10 && a <= 0xFF3F)
			return sound_read(a & 0x1FF);
		if ((a & 0xFF80) == 0xFF80)
			return Hi_Memory[a & 0x1FF];
		return ioreg_read(a & 0x1FF);
	}
	return 0xff; /* not reached */
}

void GAMEBOY::mbc_reset()
{
	mbc_rombank = 1;
	mbc_rambank = 0;
	mbc_enableram = 1;
	mbc_model = 0;
	mem_updatemap();
}



//#define C (cpu.lcdc)


/*
 * stat_trigger updates the STAT interrupt line to reflect whether any
 * of the conditions set to be tested (by bits 3-6 of R_STAT) are met.
 * This function should be called whenever any of the following occur:
 * 1) LY or LYC changes.
 * 2) A state transition affects the low 2 bits of R_STAT (see below).
 * 3) The program writes to the upper bits of R_STAT.
 * stat_trigger also updates bit 2 of R_STAT to reflect whether LY=LYC.
 */

void GAMEBOY::stat_trigger()
{
	static const int condbits[4] = { 0x08, 0x30, 0x20, 0x00 };
	int flag = 0;

	if (/*(R_LY < 0x91) &&*/ (R_LY == R_LYC))
	{
		R_STAT |= 0x04;
		if (R_STAT & 0x40) flag = IF_STAT;
	}
	else R_STAT &= ~0x04;

	if (R_STAT & condbits[R_STAT&3]) flag = IF_STAT;

	if (!(R_LCDC & 0x80)) flag = 0;
	
	hw_interrupt(flag, IF_STAT);
}

void GAMEBOY::stat_write(byte b)
{
	R_STAT = (R_STAT & 0x07) | (b & 0x78);
	if (!hw.cgb && !(R_STAT & 2)) /* DMG STAT write bug => interrupt */
		hw_interrupt(IF_STAT, IF_STAT);
	stat_trigger();
}


/*
 * stat_change is called when a transition results in a change to the
 * LCD STAT condition (the low 2 bits of R_STAT).  It raises or lowers
 * the VBLANK interrupt line appropriately and calls stat_trigger to
 * update the STAT interrupt line.
 */

void GAMEBOY::stat_change(int stat)
{
	stat &= 3;
	R_STAT = (R_STAT & 0x7C) | stat;

	if (stat != 1) hw_interrupt(0, IF_VBLANK);
	/* hw_interrupt((stat == 1) ? IF_VBLANK : 0, IF_VBLANK); */
	stat_trigger();
}


void GAMEBOY::lcdc_change(byte b)
{
	byte old = R_LCDC;
	R_LCDC = b;
	if ((R_LCDC ^ old) & 0x80) /* lcd on/off change */
	{
		R_LY = 0;
		stat_change(2);
		gb_cpu.lcdc = 40;
		lcd_begin();
	}
}


void GAMEBOY::lcdc_trans()
{
	if (!(R_LCDC & 0x80))
	{
		while (gb_cpu.lcdc  <= 0)
		{
			switch ((byte)(R_STAT & 3))
			{
			case 0:
			case 1:
				stat_change(2);
				gb_cpu.lcdc  += 40;
				break;
			case 2:
				stat_change(3);
				gb_cpu.lcdc  += 86;
				break;
			case 3:
				stat_change(0);
				if (hw.hdma & 0x80)
					hw_hdma();
				else
					gb_cpu.lcdc  += 102;
				break;
			}
			return;
		}
	}
	while (gb_cpu.lcdc  <= 0)
	{
		switch ((byte)(R_STAT & 3))
		{
		case 1:
			if (!(hw.ilines & IF_VBLANK))
			{
				gb_cpu.lcdc  += 218;
				hw_interrupt(IF_VBLANK, IF_VBLANK);
				break;
			}
			if (R_LY == 0)
			{
				lcd_begin();
				stat_change(2);
				gb_cpu.lcdc  += 40;
				break;
			}
			else if (R_LY < 152)
				gb_cpu.lcdc  += 228;
			else if (R_LY == 152)
				gb_cpu.lcdc  += 28;
			else
			{
				R_LY = -1;
				gb_cpu.lcdc  += 200;
			}
			R_LY++;
			stat_trigger();
			break;
		case 2:
			lcd_refreshline();
			stat_change(3);
			gb_cpu.lcdc  += 86;
			break;
		case 3:
			stat_change(0);
			if (hw.hdma & 0x80)
				hw_hdma();
			/* FIXME -- how much of the hblank does hdma use?? */
			/* else */
			gb_cpu.lcdc  += 102;
			break;
		case 0:
			if (++R_LY >= 144)
			{
				if (gb_cpu.halt)
				{
					hw_interrupt(IF_VBLANK, IF_VBLANK);
					gb_cpu.lcdc += 228;
				}
				else gb_cpu.lcdc += 10;
				stat_change(1);
				break;
			}
			stat_change(2);
			gb_cpu.lcdc  += 40;
			break;
		}
	}
}



#ifdef USE_ASM
#include "asm.h"
#endif

#define BG (scan.bg)
#define WND (scan.wnd)
#define BUF (scan.buf)
#define PRI (scan.pri)

#define PAL1 (scan.pal1)
#define PAL2 (scan.pal2)
#define PAL4 (scan.pal4)

#define VS (scan.vs) /* vissprites */
#define NS (scan.ns)

#define L (scan.l) /* line */
#define X (scan.x) /* screen position */
#define Y (scan.y)
#define S (scan.s) /* tilemap position */
#define T (scan.t)
#define U (scan.u) /* position within tile */
#define V (scan.v)
#define WX (scan.wx)
#define WY (scan.wy)
#define WT (scan.wt)
#define WV (scan.wv)


//#define DEF_PAL { 0x98d0e0, 0x68a0b0, 0x60707C, 0x2C3C3C }
#define DEF_PAL { 0xffffff, 0x808080, 0x232323, 0x000000 }


static int dmg_pal[4][4] = { DEF_PAL, DEF_PAL, DEF_PAL, DEF_PAL };

static int usefilter, filterdmg;
static int filter[3][4] = {
	{ 195,  25,   0,  35 },
	{  25, 170,  25,  35 },
	{  25,  60, 125,  40 }
};



#ifdef ALLOW_UNALIGNED_IO /* long long is ok since this is i386-only anyway? */
#define MEMCPY8(d, s) ((*(long long *)(d)) = (*(long long *)(s)))
#else
#define MEMCPY8(d, s) memcpy((d), (s), 8)
#endif




#ifndef ASM_UPDATEPATPIX
void GAMEBOY::updatepatpix()
{
	int i, j, k;
	int a, c;
	byte *vram = lcd.vbank[0];

	if (!anydirty) return;
	for (i = 0; i < 1024; i++)
	{
		if (i == 384) i = 512;
		if (i == 896) break;
		if (!patdirty[i]) continue;
		patdirty[i] = 0;
		for (j = 0; j < 8; j++)
		{
			a = ((i<<4) | (j<<1));
			for (k = 0; k < 8; k++)
			{
				c = vram[a] & (1<<k) ? 1 : 0;
				c |= vram[a+1] & (1<<k) ? 2 : 0;
				patpix[i+1024][j][k] = c;
			}
			for (k = 0; k < 8; k++)
				patpix[i][j][k] =
					patpix[i+1024][j][7-k];
		}
		for (j = 0; j < 8; j++)
		{
			for (k = 0; k < 8; k++)
			{
				patpix[i+2048][j][k] =
					patpix[i][7-j][k];
				patpix[i+3072][j][k] =
					patpix[i+1024][7-j][k];
			}
		}
	}
	anydirty = 0;
}
#endif /* ASM_UPDATEPATPIX */



void GAMEBOY::tilebuf()
{
	int i, cnt;
	int base;
	byte *tilemap, *attrmap;
	int *tilebuf;
	int *wrap;
	static int wraptable[64] =
	{
		0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,-32
	};

	base = ((R_LCDC&0x08)?0x1C00:0x1800) + (T<<5) + S;
	tilemap = lcd.vbank[0] + base;
	attrmap = lcd.vbank[1] + base;
	tilebuf = BG;
	wrap = wraptable + S;
	cnt = ((WX + 7) >> 3) + 1;

	if (hw.cgb)
	{
		if (R_LCDC & 0x10)
			for (i = cnt; i > 0; i--)
			{
				*(tilebuf++) = *tilemap
					| (((int)*attrmap & 0x08) << 6)
					| (((int)*attrmap & 0x60) << 5);
				*(tilebuf++) = (((int)*attrmap & 0x07) << 2);
				attrmap += *wrap + 1;
				tilemap += *(wrap++) + 1;
			}
		else
			for (i = cnt; i > 0; i--)
			{
				*(tilebuf++) = (256 + ((n8)*tilemap))
					| (((int)*attrmap & 0x08) << 6)
					| (((int)*attrmap & 0x60) << 5);
				*(tilebuf++) = (((int)*attrmap & 0x07) << 2);
				attrmap += *wrap + 1;
				tilemap += *(wrap++) + 1;
			}
	}
	else
	{
		if (R_LCDC & 0x10)
			for (i = cnt; i > 0; i--)
			{
				*(tilebuf++) = *(tilemap++);
				tilemap += *(wrap++);
			}
		else
			for (i = cnt; i > 0; i--)
			{
				*(tilebuf++) = (256 + ((n8)*(tilemap++)));
				tilemap += *(wrap++);
			}
	}

	if (WX >= 160) return;
	
	base = ((R_LCDC&0x40)?0x1C00:0x1800) + (WT<<5);
	tilemap = lcd.vbank[0] + base;
	attrmap = lcd.vbank[1] + base;
	tilebuf = WND;
	cnt = ((160 - WX) >> 3) + 1;

	if (hw.cgb)
	{
		if (R_LCDC & 0x10)
			for (i = cnt; i > 0; i--)
			{
				*(tilebuf++) = *(tilemap++)
					| (((int)*attrmap & 0x08) << 6)
					| (((int)*attrmap & 0x60) << 5);
				*(tilebuf++) = (((int)*(attrmap++)&7) << 2);
			}
		else
			for (i = cnt; i > 0; i--)
			{
				*(tilebuf++) = (256 + ((n8)*(tilemap++)))
					| (((int)*attrmap & 0x08) << 6)
					| (((int)*attrmap & 0x60) << 5);
				*(tilebuf++) = (((int)*(attrmap++)&7) << 2);
			}
	}
	else
	{
		if (R_LCDC & 0x10)
			for (i = cnt; i > 0; i--)
				*(tilebuf++) = *(tilemap++);
		else
			for (i = cnt; i > 0; i--)
				*(tilebuf++) = (256 + ((n8)*(tilemap++)));
	}
}


void GAMEBOY::bg_scan()
{
	int cnt;
	byte *src, *dest;
	int *tile;

	if (WX <= 0) return;
	cnt = WX;
	tile = BG;
	dest = BUF;
	
	src = patpix[*(tile++)][V] + U;
	memcpy(dest, src, 8-U);
	dest += 8-U;
	cnt -= 8-U;
	if (cnt <= 0) return;
	while (cnt >= 8)
	{
		src = patpix[*(tile++)][V];
		MEMCPY8(dest, src);
		dest += 8;
		cnt -= 8;
	}
	src = patpix[*tile][V];
	while (cnt--)
		*(dest++) = *(src++);
}

void GAMEBOY::wnd_scan()
{
	int cnt;
	byte *src, *dest;
	int *tile;

	if (WX >= 160) return;
	cnt = 160 - WX;
	tile = WND;
	dest = BUF + WX;
	
	while (cnt >= 8)
	{
		src = patpix[*(tile++)][WV];
		MEMCPY8(dest, src);
		dest += 8;
		cnt -= 8;
	}
	src = patpix[*tile][WV];
	while (cnt--)
		*(dest++) = *(src++);
}

void GAMEBOY::blendcpy(byte *dest, byte *src, byte b, int cnt)
{
	while (cnt--) *(dest++) = *(src++) | b;
}

int GAMEBOY::priused(void *attr)
{
	un32 *a = (un32 *)attr;
	return (int)((a[0]|a[1]|a[2]|a[3]|a[4]|a[5]|a[6]|a[7])&0x80808080);
}

void GAMEBOY::bg_scan_pri()
{
	int cnt, i;
	byte *src, *dest;

	if (WX <= 0) return;
	i = S;
	cnt = WX;
	dest = PRI;
	src = lcd.vbank[1] + ((R_LCDC&0x08)?0x1C00:0x1800) + (T<<5);

	if (!priused(src))
	{
		memset(dest, 0, cnt);
		return;
	}
	
	memset(dest, src[i++&31]&128, 8-U);
	dest += 8-U;
	cnt -= 8-U;
	if (cnt <= 0) return;
	while (cnt >= 8)
	{
		memset(dest, src[i++&31]&128, 8);
		dest += 8;
		cnt -= 8;
	}
	memset(dest, src[i&31]&128, cnt);
}

void GAMEBOY::wnd_scan_pri()
{
	int cnt, i;
	byte *src, *dest;

	if (WX >= 160) return;
	i = 0;
	cnt = 160 - WX;
	dest = PRI + WX;
	src = lcd.vbank[1] + ((R_LCDC&0x40)?0x1C00:0x1800) + (WT<<5);
	
	if (!priused(src))
	{
		memset(dest, 0, cnt);
		return;
	}
	
	while (cnt >= 8)
	{
		memset(dest, src[i++]&128, 8);
		dest += 8;
		cnt -= 8;
	}
	memset(dest, src[i]&128, cnt);
}

#ifndef ASM_BG_SCAN_COLOR
void GAMEBOY::bg_scan_color()
{
	int cnt;
	byte *src, *dest;
	int *tile;

	if (WX <= 0) return;
	cnt = WX;
	tile = BG;
	dest = BUF;
	
	src = patpix[*(tile++)][V] + U;
	blendcpy(dest, src, *(tile++), 8-U);
	dest += 8-U;
	cnt -= 8-U;
	if (cnt <= 0) return;
	while (cnt >= 8)
	{
		src = patpix[*(tile++)][V];
		blendcpy(dest, src, *(tile++), 8);
		dest += 8;
		cnt -= 8;
	}
	src = patpix[*(tile++)][V];
	blendcpy(dest, src, *(tile++), cnt);
}
#endif

void GAMEBOY::wnd_scan_color()
{
	int cnt;
	byte *src, *dest;
	int *tile;

	if (WX >= 160) return;
	cnt = 160 - WX;
	tile = WND;
	dest = BUF + WX;
	
	while (cnt >= 8)
	{
		src = patpix[*(tile++)][WV];
		blendcpy(dest, src, *(tile++), 8);
		dest += 8;
		cnt -= 8;
	}
	src = patpix[*(tile++)][WV];
	blendcpy(dest, src, *(tile++), cnt);
}

void GAMEBOY::recolor(byte *buf, byte fill, int cnt)
{
	while (cnt--) *(buf++) |= fill;
}

void GAMEBOY::spr_count()
{
	int i;
	struct obj *o;
	
	NS = 0;
	if (!(R_LCDC & 0x02)) return;
	
	o = lcd.oam.obj;
	
	for (i = 40; i; i--, o++)
	{
		if (L >= o->y || L + 16 < o->y)
			continue;
		if (L + 8 >= o->y && !(R_LCDC & 0x04))
			continue;
		if (++NS == 10) break;
	}
}

void GAMEBOY::spr_enum()
{
	int i, j;
	struct obj *o;
	struct vissprite ts[10];
	int v, pat;
	int l, x;

	NS = 0;
	if (!(R_LCDC & 0x02)) return;

	o = lcd.oam.obj;
	
	for (i = 40; i; i--, o++)
	{
		if (L >= o->y || L + 16 < o->y)
			continue;
		if (L + 8 >= o->y && !(R_LCDC & 0x04))
			continue;
		VS[NS].x = (int)o->x - 8;
		v = L - (int)o->y + 16;
		if (hw.cgb)
		{
			pat = o->pat | (((int)o->flags & 0x60) << 5)
				| (((int)o->flags & 0x08) << 6);
			VS[NS].pal = 32 + ((o->flags & 0x07) << 2);
		}
		else
		{
			pat = o->pat | (((int)o->flags & 0x60) << 5);
			VS[NS].pal = 32 + ((o->flags & 0x10) >> 2);
		}
		VS[NS].pri = (o->flags & 0x80) >> 7;
		if ((R_LCDC & 0x04))
		{
			pat &= ~1;
			if (v >= 8)
			{
				v -= 8;
				pat++;
			}
			if (o->flags & 0x40) pat ^= 1;
		}
		VS[NS].buf = patpix[pat][v];
		if (++NS == 10) break;
	}
	if (!sprsort || hw.cgb) return;
	/* not quite optimal but it finally works! */
	for (i = 0; i < NS; i++)
	{
		l = 0;
		x = VS[0].x;
		for (j = 1; j < NS; j++)
		{
			if (VS[j].x < x)
			{
				l = j;
				x = VS[j].x;
			}
		}
		ts[i] = VS[l];
		VS[l].x = 160;
	}
	memcpy(VS, ts, sizeof VS);
}

void GAMEBOY::spr_scan()
{
	int i, x;
	byte pal, b, ns = NS;
	byte *src, *dest, *bg, *pri;
	struct vissprite *vs;
	static byte bgdup[256];

	if (!ns) return;

	memcpy(bgdup, BUF, 256);
	vs = &VS[ns-1];
	
	for (; ns; ns--, vs--)
	{
		x = vs->x;
		if (x >= 160) continue;
		if (x <= -8) continue;
		if (x < 0)
		{
			src = vs->buf - x;
			dest = BUF;
			i = 8 + x;
		}
		else
		{
			src = vs->buf;
			dest = BUF + x;
			if (x > 152) i = 160 - x;
			else i = 8;
		}
		pal = vs->pal;
		if (vs->pri)
		{
			bg = bgdup + (dest - BUF);
			while (i--)
			{
				b = src[i];
				if (b && !(bg[i]&3)) dest[i] = pal|b;
			}
		}
		else if (hw.cgb)
		{
			bg = bgdup + (dest - BUF);
			pri = PRI + (dest - BUF);
			while (i--)
			{
				b = src[i];
				if (b && (!pri[i] || !(bg[i]&3)))
					dest[i] = pal|b;
			}
		}
		else while (i--) if (src[i]) dest[i] = pal|src[i];
		/* else while (i--) if (src[i]) dest[i] = 31 + ns; */
	}
	if (sprdebug) for (i = 0; i < NS; i++) BUF[i<<1] = 36;
}






void GAMEBOY::lcd_begin()
{
	if (fb.indexed)
	{
		if (rgb332) pal_set332();
		else pal_expire();
	}
//	while (scale * 160 > fb.w || scale * 144 > fb.h) scale--;
	vdest = fb.ptr;
/*
	vdest = fb.ptr + ((fb.w*fb.pelsize)>>1)
		- (80*fb.pelsize) * scale
		+ ((fb.h>>1) - 72*scale) * fb.pitch;
*/
	WY = R_WY;
}

void GAMEBOY::lcd_refreshline()
{
	int i;
	byte scalebuf[160*4*4], *dest;
	
	if (!fb.enabled) return;
	
	if (!(R_LCDC & 0x80))
		return; /* should not happen... */
	
	updatepatpix();

	L = R_LY;
	X = R_SCX;
	Y = (R_SCY + L) & 0xff;
	S = X >> 3;
	T = Y >> 3;
	U = X & 7;
	V = Y & 7;
	
	WX = R_WX - 7;
	if (WY>L || WY<0 || WY>143 || WX<-7 || WX>159 || !(R_LCDC&0x20))
		WX = 160;
	WT = (L - WY) >> 3;
	WV = (L - WY) & 7;

	spr_enum();

	tilebuf();
	if (hw.cgb)
	{
		bg_scan_color();
		wnd_scan_color();
		if (NS)
		{
			bg_scan_pri();
			wnd_scan_pri();
		}
	}
	else
	{
		bg_scan();
		wnd_scan();
		recolor(BUF+WX, 0x04, 160-WX);
	}
	spr_scan();

	if (fb.dirty) memset(fb.ptr, 0, fb.pitch * fb.h);
	fb.dirty = 0;
	if (density > scale) density = scale;
	if (scale == 1) density = 1;

	dest = (density != 1) ? scalebuf : vdest;
	
	switch (scale)
	{
	case 0:
	case 1:
		switch (fb.pelsize)
		{
		case 1:
			refresh_1(dest, BUF, PAL1, 160);
			break;
		case 2:
			refresh_2((un16 *)dest, BUF, PAL2, 160);
			break;
		case 3:
			refresh_3(dest, BUF, (un32 *)PAL4, 160);
			break;
		case 4:
			refresh_4((un32 *)dest, BUF, (un32 *)PAL4, 160);
			break;
		}
		break;
	case 2:
		switch (fb.pelsize)
		{
		case 1:
			refresh_2((un16 *)dest, BUF, PAL2, 160);
			break;
		case 2:
			refresh_4((un32 *)dest, BUF, (un32 *)PAL4, 160);
			break;
		case 3:
			refresh_3_2x(dest, BUF, PAL4, 160);
			break;
		case 4:
			refresh_4_2x((un32 *)dest, BUF, PAL4, 160);
			break;
		}
		break;
	case 3:
		switch (fb.pelsize)
		{
		case 1:
			refresh_3(dest, BUF, (un32 *)PAL4, 160);
			break;
		case 2:
			refresh_2_3x((un16 *)dest, BUF, PAL2, 160);
			break;
		case 3:
			refresh_3_3x(dest, BUF, PAL4, 160);
			break;
		case 4:
			refresh_4_3x((un32 *)dest, BUF, PAL4, 160);
			break;
		}
		break;
	case 4:
		switch (fb.pelsize)
		{
		case 1:
			refresh_4((un32 *)dest, BUF, (un32 *)PAL4, 160);
			break;
		case 2:
			refresh_4_2x((un32 *)dest, BUF, PAL4, 160);
			break;
		case 3:
			refresh_3_4x(dest, BUF, PAL4, 160);
			break;
		case 4:
			refresh_4_4x((un32 *)dest, BUF, PAL4, 160);
			break;
		}
		break;
	default:
		break;
	}

	if (density != 1)
	{
		for (i = 0; i < scale; i++)
		{
			if ((i < density) || ((density <= 0) && !(i&1)))
				memcpy(vdest, scalebuf, 160 * fb.pelsize * scale);
			vdest += fb.pitch;
		}
	}
	else vdest += fb.pitch * scale;
}







void GAMEBOY::updatepalette(int i)
{
	int c, r, g, b, y, u, v, rr, gg;

	c = (lcd.pal[i<<1] | ((int)lcd.pal[(i<<1)|1] << 8)); // & 0x7FFF;
	r = (c & 0x001F) << 3;
	g = (c & 0x03E0) >> 2;
	b = (c & 0x7C00) >> 7;//7;
	r |= (r >> 5);
	g |= (g >> 5);
	b |= (b >> 5);
/*
	if (usefilter && (filterdmg || hw.cgb))
	{
		rr = ((r * filter[0][0] + g * filter[0][1] + b * filter[0][2]) >> 8) + filter[0][3];
		gg = ((r * filter[1][0] + g * filter[1][1] + b * filter[1][2]) >> 8) + filter[1][3];
		b = ((r * filter[2][0] + g * filter[2][1] + b * filter[2][2]) >> 8) + filter[2][3];
		r = rr;
		g = gg;
	}

	if (fb.yuv)
	{
		y = (((r *  263) + (g * 516) + (b * 100)) >> 10) + 16;
		u = (((r *  450) - (g * 377) - (b *  73)) >> 10) + 128;
		v = (((r * -152) - (g * 298) + (b * 450)) >> 10) + 128;
		if (y < 0) y = 0; if (y > 255) y = 255;
		if (u < 0) u = 0; if (u > 255) u = 255;
		if (v < 0) v = 0; if (v > 255) v = 255;
		PAL4[i] = (y<<fb.cc[0].l) | (y<<fb.cc[3].l)
			| (u<<fb.cc[1].l) | (v<<fb.cc[2].l);
		return;
	}

	if (fb.indexed)
	{
		pal_release(PAL1[i]);
		c = pal_getcolor(c, r, g, b);
		PAL1[i] = c;
		PAL2[i] = (c<<8) | c;
		PAL4[i] = (c<<24) | (c<<16) | (c<<8) | c;
		return;
	}
*/

	r = (r >> fb.cc[0].r) << fb.cc[0].l;
	g = (g >> fb.cc[1].r) << fb.cc[1].l;
	b = (b >> fb.cc[2].r) << fb.cc[2].l;
	c = r|g|b;
/*
	{
		// convert 555 -> 555 ( xBBBBB GGGGG RRRRR  ->  xRRRRR GGGGG BBBBB )
		r = (c & 0x001F) << 10;
		g = (c & 0x03E0);
		b = (c & 0x7C00) >> 10;
		c = r|g|b;
	}
*/
	switch (fb.pelsize)
	{
	case 1:
		PAL1[i] = c;
		PAL2[i] = (c<<8) | c;
		PAL4[i] = (c<<24) | (c<<16) | (c<<8) | c;
		break;
	case 2:
		PAL2[i] = c;
		PAL4[i] = (c<<16) | c;
		break;
	case 3:
	case 4:
		PAL4[i] = c;
		break;
	}
}

void GAMEBOY::pal_write(int i, byte b)
{
	if (lcd.pal[i] == b) return;
	lcd.pal[i] = b;
	updatepalette(i>>1);
}

void GAMEBOY::pal_write_dmg(int i, int mapnum, byte d)
{
	int j;
	int *cmap = dmg_pal[mapnum];
	int c, r, g, b;

	if (hw.cgb) return;

	/* if (mapnum >= 2) d = 0xe4; */
	for (j = 0; j < 8; j += 2)
	{
/*
		c = cmap[(d >> j) & 3];
		r = (c & 0xf8) >> 3;
		g = (c & 0xf800) >> 6;//6;
		b = (c & 0xf80000) >> 9;
*/
		unsigned char *cp = (unsigned char *)&cmap[(d >> j) & 3];

		r = (cp[0] <<7) & 0x7c00;//0xf800;
		g = (cp[1] <<2) & 0x3e0;//6;
		b = (cp[2] >>3) & 0x1f;
		c = r|g|b;
		/* FIXME - handle directly without faking cgb */
		pal_write(i+j, c & 0xff);
		pal_write(i+j+1, c >> 8);
	}
}

void GAMEBOY::vram_write(int a, byte b)
{
	lcd.vbank[R_VBK&1][a] = b;
	if (a >= 0x1800) return;
	patdirty[((R_VBK&1)<<9)+(a>>4)] = 1;
	anydirty = 1;
}

void GAMEBOY::vram_dirty()
{
	anydirty = 1;
	memset(patdirty, 1, sizeof patdirty);
}

void GAMEBOY::pal_dirty()
{
	int i;
	if (!hw.cgb)
	{
		pal_write_dmg(0, 0, R_BGP);
		pal_write_dmg(8, 1, R_BGP);
		pal_write_dmg(64, 2, R_OBP0);
		pal_write_dmg(72, 3, R_OBP1);
	}
	for (i = 0; i < 64; i++)
		updatepalette(i);
}

void GAMEBOY::lcd_reset()
{
	memset(&lcd, 0, sizeof lcd);
	lcd_begin();
	vram_dirty();
	pal_dirty();
}












/*
 * hw_interrupt changes the virtual interrupt lines included in the
 * specified mask to the values the corresponding bits in i take, and
 * in doing so, raises the appropriate bit of R_IF for any interrupt
 * lines that transition from low to high.
 */

void GAMEBOY::hw_interrupt(byte i, byte mask)
{
	byte oldif = R_IF;
	i &= 0x1F & mask;
	R_IF |= i & (hw.ilines ^ i);

	/* FIXME - is this correct? not sure the docs understand... */
	if ((R_IF & (R_IF ^ oldif) & R_IE) && gb_cpu.ime) gb_cpu.halt = 0;
	/* if ((i & (hw.ilines ^ i) & R_IE) && cpu.ime) cpu.halt = 0; */
	/* if ((i & R_IE) && cpu.ime) cpu.halt = 0; */
	
	hw.ilines &= ~mask;
	hw.ilines |= i;
}


/*
 * hw_dma performs plain old memory-to-oam dma, the original dmg
 * dma. Although on the hardware it takes a good deal of time, the cpu
 * continues running during this mode of dma, so no special tricks to
 * stall the cpu are necessary.
 */

void GAMEBOY::hw_dma(byte b)
{
	int i;
	addr a;

	a = ((addr)b) << 8;
	for (i = 0; i < 160; i++, a++)
		lcd.oam.mem[i] = mem_read(a);
}



void GAMEBOY::hw_hdma_cmd(byte c)
{
	int cnt;
	addr sa;
	int da;

	/* Begin or cancel HDMA */
	if ((hw.hdma|c) & 0x80)
	{
		hw.hdma = c;
		R_HDMA5 = c & 0x7f;
		return;
	}
	
	/* Perform GDMA */
	sa = ((addr)R_HDMA1 << 8) | (R_HDMA2&0xf0);
	da = 0x8000 | ((int)(R_HDMA3&0x1f) << 8) | (R_HDMA4&0xf0);
	cnt = ((int)c)+1;
	/* FIXME - this should use cpu time! */
	/*cpu_timers(102 * cnt);*/
	cnt <<= 4;
	while (cnt--)
		mem_write(da++, mem_read(sa++));
	R_HDMA1 = sa >> 8;
	R_HDMA2 = sa & 0xF0;
	R_HDMA3 = 0x1F & (da >> 8);
	R_HDMA4 = da & 0xF0;
	R_HDMA5 = 0xFF;
}


void GAMEBOY::hw_hdma()
{
	int cnt;
	addr sa;
	int da;

	sa = ((addr)R_HDMA1 << 8) | (R_HDMA2&0xf0);
	da = 0x8000 | ((int)(R_HDMA3&0x1f) << 8) | (R_HDMA4&0xf0);
	cnt = 16;
	while (cnt--)
		mem_write(da++, mem_read(sa++));
	R_HDMA1 = sa >> 8;
	R_HDMA2 = sa & 0xF0;
	R_HDMA3 = 0x1F & (da >> 8);
	R_HDMA4 = da & 0xF0;
	R_HDMA5--;
	hw.hdma--;
}


/*
 * pad_refresh updates the P1 register from the pad states, generating
 * the appropriate interrupts (by quickly raising and lowering the
 * interrupt line) if a transition has been made.
 */

void GAMEBOY::pad_refresh()
{
	byte oldp1;
	oldp1 = R_P1;
	R_P1 &= 0x30;
	R_P1 |= 0xc0;
	if (!(R_P1 & 0x10))
		R_P1 |= (hw.pad & 0x0F);
	if (!(R_P1 & 0x20))
		R_P1 |= (hw.pad >> 4);
	R_P1 ^= 0x0F;
	if (oldp1 & ~R_P1 & 0x0F)
	{
		hw_interrupt(IF_PAD, IF_PAD);
		hw_interrupt(0, IF_PAD);
	}
}


/*
 * These simple functions just update the state of a button on the
 * pad.
 */

void GAMEBOY::pad_check()
{
	if(*Pad != hw.pad){
		for(unsigned char c=0x80; c; c>>=1){
			if((hw.pad & c) && !(*Pad & c))				// release
				pad_release(c);
			else if(!(hw.pad & c) && (*Pad & c))		// press
				pad_press(c);
		}
		hw.pad = *Pad;
	}
}


void GAMEBOY::pad_press(byte k)
{
	if (hw.pad & k)
		return;
	hw.pad |= k;
	pad_refresh();
}

void GAMEBOY::pad_release(byte k)
{
	if (!(hw.pad & k))
		return;
	hw.pad &= ~k;
	pad_refresh();
}

void GAMEBOY::pad_set(byte k, int st)
{
	st ? pad_press(k) : pad_release(k);
}

void GAMEBOY::hw_reset()
{
	hw.ilines = hw.pad = 0;

	memset(Hi_Memory, 0, sizeof Hi_Memory);

	R_P1 = 0xFF;
	R_LCDC = 0x91;
	R_BGP = 0xFC;
	R_OBP0 = 0xFF;
	R_OBP1 = 0xFF;
	R_SVBK = 0x01;
	R_HDMA5 = 0xFF;
	R_VBK = 0xFE;
}








static int syncrtc = 1;



void GAMEBOY::rtc_latch(byte b)
{
	if ((rtc.latch ^ b) & b & 1)
	{
		rtc.regs[0] = rtc.s;
		rtc.regs[1] = rtc.m;
		rtc.regs[2] = rtc.h;
		rtc.regs[3] = rtc.d;
		rtc.regs[4] = (rtc.d>>9) | (rtc.stop<<6) | (rtc.carry<<7);
		rtc.regs[5] = 0xff;
		rtc.regs[6] = 0xff;
		rtc.regs[7] = 0xff;
	}
	rtc.latch = b;
}

void GAMEBOY::rtc_write(byte b)
{
	/* printf("write %02X: %02X (%d)\n", rtc.sel, b, b); */
	if (!(rtc.sel & 8)) return;
	switch (rtc.sel & 7)
	{
	case 0:
		rtc.s = rtc.regs[0] = b;
		while (rtc.s >= 60) rtc.s -= 60;
		break;
	case 1:
		rtc.m = rtc.regs[1] = b;
		while (rtc.m >= 60) rtc.m -= 60;
		break;
	case 2:
		rtc.h = rtc.regs[2] = b;
		while (rtc.h >= 24) rtc.h -= 24;
		break;
	case 3:
		rtc.regs[3] = b;
		rtc.d = (rtc.d & 0x100) | b;
		break;
	case 4:
		rtc.regs[4] = b;
		rtc.d = (rtc.d & 0xff) | ((b&1)<<9);
		rtc.stop = (b>>6)&1;
		rtc.carry = (b>>7)&1;
		break;
	}
}

void GAMEBOY::rtc_tick()
{
	if (rtc.stop) return;
	if (++rtc.t == 60)
	{
		if (++rtc.s == 60)
		{
			if (++rtc.m == 60)
			{
				if (++rtc.h == 24)
				{
					if (++rtc.d == 365)
					{
						rtc.d = 0;
						rtc.carry = 1;
					}
					rtc.h = 0;
				}
				rtc.m = 0;
			}
			rtc.s = 0;
		}
		rtc.t = 0;
	}
}

void GAMEBOY::rtc_save_internal(FILE *f)
{
	fprintf(f, "%d %d %d %02d %02d %02d %02d\n%d\n",
		rtc.carry, rtc.stop, rtc.d, rtc.h, rtc.m, rtc.s, rtc.t,
		time(0));
}

void GAMEBOY::rtc_load_internal(FILE *f)
{
	int rt = 0;
	fscanf(
		f, "%d %d %d %02d %02d %02d %02d\n%d\n",
		&rtc.carry, &rtc.stop, &rtc.d,
		&rtc.h, &rtc.m, &rtc.s, &rtc.t, &rt);
	while (rtc.t >= 60) rtc.t -= 60;
	while (rtc.s >= 60) rtc.s -= 60;
	while (rtc.m >= 60) rtc.m -= 60;
	while (rtc.h >= 24) rtc.h -= 24;
	while (rtc.d >= 365) rtc.d -= 365;
	rtc.stop &= 1;
	rtc.carry &= 1;
	if (rt) rt = (time(0) - rt) * 60;
	if (syncrtc) while (rt-- > 0) rtc_tick();
}





byte GAMEBOY::bestmatch(int c)
{
	byte n, best;
	int r, g, b;
	int r2, g2, b2, c2;
	int err, besterr;

	r = (c & 0x001F) << 3;
	g = (c & 0x03E0) >> 2;
	b = (c & 0x7C00) >> 7;

	best = 0;
	besterr = 1024;
	for (n = 1; n; n++)
	{
		c2 = palrev[n];
		r2 = (c2 & 0x001F) << 3;
		g2 = (c2 & 0x03E0) >> 2;
		b2 = (c2 & 0x7C00) >> 7;
		err = abs(r-r2) + abs(b-b2) + abs(g-g2);
		if (err < besterr)
		{
			besterr = err;
			best = n;
		}
	}
	return best;
}

void GAMEBOY::makecourse(int c, byte n)
{
	int i;
	for (i = 0; i < 4; i++)
	{
		c &= crsmask[i];
		crsmap[i][c] = n;
		crsrev[i][n] = c;
	}
}

byte GAMEBOY::findcourse(int c)
{
	int i;
	byte n;
	for (i = 0; i < 4; i++)
	{
		c &= crsmask[i];
		n = crsmap[i][c];
		if (crsrev[i][n] == c)
			return n;
	}
	return 0;
}


void GAMEBOY::pal_lock(byte n)
{
	if (!n) return;
	if (pallock[n] >= pl_locked)
		pallock[n]++;
	else pallock[n] = pl_locked;
}

byte GAMEBOY::pal_getcolor(int c, int r, int g, int b)
{
	byte n;
	static byte l;
	n = palmap[c];
	if (n && pallock[n] && palrev[n] == c)
	{
		pal_lock(n);
		return n;
	}
	for (n = l+1; n != l; n++)
	{
		if (!n || pallock[n] /* || n < 16 */) continue;
		pal_lock(n);
		palmap[c] = n;
		palrev[n] = c;
		makecourse(c, n);
//--		vid_setpal(n, r, g, b);
		return (l = n);
	}
	n = findcourse(c);
	pal_lock(n);
	return n;
}

void GAMEBOY::pal_release(byte n)
{
	if (pallock[n] >= pl_locked)
		pallock[n]--;
}


void GAMEBOY::pal_expire()
{
	int i;
	for (i = 0; i < 256; i++)
		if (pallock[i] && pallock[i] < pl_locked)
			pallock[i]--;
}


void GAMEBOY::pal_set332()
{
	int i, r, g, b;
	
	fb.indexed = 0;
	fb.cc[0].r = 5;
	fb.cc[1].r = 5;
	fb.cc[2].r = 6;
	fb.cc[0].l = 0;
	fb.cc[1].l = 3;
	fb.cc[2].l = 6;
	
	i = 0;
//	for (b = 0; b < 4; b++) for (g = 0; g < 8; g++) for (r = 0; r < 8; r++)
//		vid_setpal(i++, (r<<5)|(r<<2)|(r>>1), (g<<5)|(g<<2)|(g>>1), (b<<6)|(b<<4)|(b<<2)|b);
}



void GAMEBOY::new_snd_mgr(sound_mgr* _sound_mgr)
{
	snd_mgr = _sound_mgr;
	pcm.hz = snd_mgr->get_sample_rate();
	pcm.len = snd_mgr->get_buffer_len()/2;
	if(pcm.len==0)
		return;
	pcm.pos = 0;
//	currently_playing_half = SOUND_BUF_LOW;
	if (pcm.hz) snd.rate = (1<<21) / pcm.hz;	// 0x200000 / rate
	else snd.rate = 0;
}





// sound

#include "GB_noise.h"


const static byte dmgwave[16] =
{
	0xac, 0xdd, 0xda, 0x48,
	0x36, 0x02, 0xcf, 0x16,
	0x2c, 0x04, 0xe5, 0x2c,
	0xac, 0xdd, 0xda, 0x48
};

const static byte cgbwave[16] =
{
	0x00, 0xff, 0x00, 0xff,
	0x00, 0xff, 0x00, 0xff,
	0x00, 0xff, 0x00, 0xff,
	0x00, 0xff, 0x00, 0xff,
};

const static byte sqwave[4][8] =
{
	{  0, 0,-1, 0, 0, 0, 0, 0 },
	{  0,-1,-1, 0, 0, 0, 0, 0 },
	{ -1,-1,-1,-1, 0, 0, 0, 0 },
	{ -1, 0, 0,-1,-1,-1,-1,-1 }
};

const static int freqtab[8] =
{
	(1<<14)*2,
	(1<<14),
	(1<<14)/2,
	(1<<14)/3,
	(1<<14)/4,
	(1<<14)/5,
	(1<<14)/6,
	(1<<14)/7
};


#define RATE (snd.rate)
#define WAVE (snd.wave) /* ram.hi+0x30 */
#define S1 (snd.ch[0])
#define S2 (snd.ch[1])
#define S3 (snd.ch[2])
#define S4 (snd.ch[3])



void GAMEBOY::s1_freq_d(int d)
{
	if (RATE > (d<<4)) S1.freq = 0;
	else S1.freq = (RATE << 17)/d;
}

void GAMEBOY::s1_freq()
{
	s1_freq_d(2048 - (((R_NR14&7)<<8) + R_NR13));
}

void GAMEBOY::s2_freq()
{
	int d = 2048 - (((R_NR24&7)<<8) + R_NR23);
	if (RATE > (d<<4)) S2.freq = 0;
	else S2.freq = (RATE << 17)/d;
}

void GAMEBOY::s3_freq()
{
	int d = 2048 - (((R_NR34&7)<<8) + R_NR33);
	if (RATE > (d<<3)) S3.freq = 0;
	else S3.freq = (RATE << 21)/d;
}

void GAMEBOY::s4_freq()
{
	S4.freq = (freqtab[R_NR43&7] >> (R_NR43 >> 4)) * RATE;
	if (S4.freq >> 18) S4.freq = 1<<18;
}

void GAMEBOY::sound_dirty()
{
	S1.swlen = ((R_NR10>>4) & 7) << 14;
	S1.len = (64-(R_NR11&63)) << 13;
	S1.envol = R_NR12 >> 4;
	S1.endir = (R_NR12>>3) & 1;
	S1.endir |= S1.endir - 1;
	S1.enlen = (R_NR12 & 7) << 15;
	s1_freq();
	S2.len = (64-(R_NR21&63)) << 13;
	S2.envol = R_NR22 >> 4;
	S2.endir = (R_NR22>>3) & 1;
	S2.endir |= S2.endir - 1;
	S2.enlen = (R_NR22 & 7) << 15;
	s2_freq();
	S3.len = (256-R_NR31) << 20;
	s3_freq();
	S4.len = (64-(R_NR41&63)) << 13;
	S4.envol = R_NR42 >> 4;
	S4.endir = (R_NR42>>3) & 1;
	S4.endir |= S4.endir - 1;
	S4.enlen = (R_NR42 & 7) << 15;
	s4_freq();
}

void GAMEBOY::sound_off()
{
	memset(&S1, 0, sizeof S1);
	memset(&S2, 0, sizeof S2);
	memset(&S3, 0, sizeof S3);
	memset(&S4, 0, sizeof S4);
	R_NR10 = 0x80;
	R_NR11 = 0xBF;
	R_NR12 = 0xF3;
	R_NR14 = 0xBF;
	R_NR21 = 0x3F;
	R_NR22 = 0x00;
	R_NR24 = 0xBF;
	R_NR30 = 0x7F;
	R_NR31 = 0xFF;
	R_NR32 = 0x9F;
	R_NR33 = 0xBF;
	R_NR41 = 0xFF;
	R_NR42 = 0x00;
	R_NR43 = 0x00;
	R_NR44 = 0xBF;
	R_NR50 = 0x77;
	R_NR51 = 0xF3;
	R_NR52 = 0xF1;
	sound_dirty();
}

void GAMEBOY::sound_reset()
{
	pcm.hz = snd_mgr->get_sample_rate();
	pcm.len = snd_mgr->get_buffer_len()/2;
	if(pcm.len==0)
		return;
	pcm.stereo = 1;
	pcm.pos = 0;
	currently_playing_half = SOUND_BUF_LOW;

	apu_qreadp = apu_qwritep = 0;

	memset(&snd, 0, sizeof snd);
	if (pcm.hz) snd.rate = (1<<21) / pcm.hz;	// 0x200000 / rate
	else snd.rate = 0;
#ifdef _DEBUG
//	LOG("sound rate " << HEX(snd.rate,8) << "sound buff len " << HEX(pcm.len,8)<<endl);
//	gb_sndclock = 0;
#endif
	snd.ch[0].on = 1;
	snd.ch[1].on = 1;
	snd.ch[2].on = 1;
	snd.ch[3].on = 1;
	memcpy(WAVE, hw.cgb ? cgbwave : dmgwave, 16);
	memcpy(Hi_Memory+0x130, WAVE, 16);
	sound_off();
#ifdef GB_SOUND_GNUBOY
#else
//	sndRegChange = 0;
	snd1Enable=snd1Freq=snd1Vol=snd1Len=snd1SwpCnt=0;
	snd1EnvCnt=snd1Bit=snd1ValL=snd1ValR=snd1Cnt=0;
	snd2Enable=snd2Freq=snd2Vol=snd2Len=0;
	snd2EnvCnt=snd2Bit=snd2ValL=snd2ValR=snd2Cnt=0;
	snd3Enable=snd3Freq=snd3Len=snd3Bit=snd3Cnt=0;
	snd3ValL=snd3ValR=0;
	snd4Enable=snd4Freq=snd4Vol=snd4Len;
	snd4EnvCnt=snd4Bit=snd4ValL=snd4ValR=snd4Cnt=0;
//sndRegChange=0;
	swpClk=0;
	envClk=0;
	lenClk=0;
	freqClk=0;
	freq4Clk=0;
#endif
}


void GAMEBOY::sound_mix()
{
	int s, l, r, f, n;

//	if (!RATE || gb_cpu.snd < RATE) return;
#ifdef GB_SOUND_GNUBOY
//	for (; gb_cpu.snd >= RATE; gb_cpu.snd -= RATE)
	{
		l = r = 0;

		if (S1.on)
		{
			s = sqwave[R_NR11>>6][(S1.pos>>18)&7] & S1.envol;
			S1.pos += S1.freq;
			if ((R_NR14 & 64) && ((S1.cnt += RATE) >= S1.len))
				S1.on = 0;
			if (S1.enlen && (S1.encnt += RATE) >= S1.enlen)
			{
				S1.encnt -= S1.enlen;
				S1.envol += S1.endir;
				if (S1.envol < 0) S1.envol = 0;
				if (S1.envol > 15) S1.envol = 15;
			}
			if (S1.swlen && (S1.swcnt += RATE) >= S1.swlen)
			{
				S1.swcnt -= S1.swlen;
				f = S1.swfreq;
				n = (R_NR10 & 7);
				if (R_NR10 & 8) f -= (f >> n);
				else f += (f >> n);
				if (f > 2047)
					S1.on = 0;
				else
				{
					S1.swfreq = f;
					R_NR13 = f;
					R_NR14 = (R_NR14 & 0xF8) | (f>>8);
					s1_freq_d(2048 - f);
				}
			}
			s <<= 2;
			if (R_NR51 & 1) r += s;
			if (R_NR51 & 16) l += s;
		}
		
		if (S2.on)
		{
			s = sqwave[R_NR21>>6][(S2.pos>>18)&7] & S2.envol;
			S2.pos += S2.freq;
			if ((R_NR24 & 64) && ((S2.cnt += RATE) >= S2.len))
				S2.on = 0;
			if (S2.enlen && (S2.encnt += RATE) >= S2.enlen)
			{
				S2.encnt -= S2.enlen;
				S2.envol += S2.endir;
				if (S2.envol < 0) S2.envol = 0;
				if (S2.envol > 15) S2.envol = 15;
			}
			s <<= 2;
			if (R_NR51 & 2) r += s;
			if (R_NR51 & 32) l += s;
		}
		
		if (S3.on)
		{
			s = WAVE[(S3.pos>>22) & 15];
			if (S3.pos & (1<<21)) s &= 15;
			else s >>= 4;
			s -= 8;
			S3.pos += S3.freq;
			if ((R_NR34 & 64) && ((S3.cnt += RATE) >= S3.len))
				S3.on = 0;
			if (R_NR32 & 96) s <<= (3 - ((R_NR32>>5)&3));
			else s = 0;
			if (R_NR51 & 4) r += s;
			if (R_NR51 & 64) l += s;
		}

		if (S4.on)
		{
			if (R_NR43 & 8) s = 1 & (noise7[(S4.pos>>20)&15] >> (7-((S4.pos>>17)&7)));
			else s = 1 & (noise15[(S4.pos>>20)&4095] >> (7-((S4.pos>>17)&7)));
			s = (-s) & S4.envol;
			S4.pos += S4.freq;
			if ((R_NR44 & 64) && ((S4.cnt += RATE) >= S4.len))
				S4.on = 0;
			if (S4.enlen && (S4.encnt += RATE) >= S4.enlen)
			{
				S4.encnt -= S4.enlen;
				S4.envol += S4.endir;
				if (S4.envol < 0) S4.envol = 0;
				if (S4.envol > 15) S4.envol = 15;
			}
			s += s << 1;
			if (R_NR51 & 8) r += s;
			if (R_NR51 & 128) l += s;
		}
		
		l *= (R_NR50 & 0x07);
		r *= ((R_NR50 & 0x70)>>4);
		l >>= 4;
		r >>= 4;
		
		if (l > 127) l = 127;
		else if (l < -128) l = -128;
		if (r > 127) r = 127;
		else if (r < -128) r = -128;
/*
		if (pcm.buf)
		{
			if(pcm.pos >= pcm.len)
				pcm_submit();
			if (pcm.stereo)
			{
				pcm.buf[pcm.pos++] = l+128;
				pcm.buf[pcm.pos++] = r+128;
			}
			else pcm.buf[pcm.pos++] = ((l+r)>>1)+128;
		}
*/
	}
	R_NR52 = (R_NR52&0xf0) | S1.on | (S2.on<<1) | (S3.on<<2) | (S4.on<<3);
#else
#endif
}


void GAMEBOY::pcm_submit(){
//	if (!pcm.buf) return;
//	if (pcm.pos < pcm.len) return;
	pcm.pos = 0;
    if(!snd_mgr->IsNull())
    {
		sound_buf_pos cur_ph;  // cur playing half
		sound_buf_pos cur_nph; // cur not-playing half
		uint8* buf;
		uint32 buf_len;

		cur_ph = snd_mgr->get_currently_playing_half();
		if(cur_ph == currently_playing_half) return;
		cur_nph = currently_playing_half;
		currently_playing_half = cur_ph;
		if(!snd_mgr->lock(cur_nph, (void**)&buf, &buf_len)){
//			LOG("couldn't lock sound buffer" << endl);
			return;
		}
#ifdef _DEBUG
//		LOG("lock sound buffer " << HEX(buf_len,8) <<endl);
//		gb_sndclock = 0;
#endif
//		memcpy(buf, pcm.buf, buf_len);
		snd_mgr->unlock();
    }
}


byte GAMEBOY::sound_read(int r)
{
//	sound_mix();
	/* printf("read %02X: %02X\n", r, REG(r)); */
	return REG(r);
}

void GAMEBOY::s1_init()
{
	S1.swcnt = 0;
	S1.swfreq = ((R_NR14&7)<<8) + R_NR13;
	S1.envol = R_NR12 >> 4;
	S1.endir = (R_NR12>>3) & 1;
	S1.endir |= S1.endir - 1;
	S1.enlen = (R_NR12 & 7) << 15;
	if (!S1.on) S1.pos = 0;
	S1.on = 1;
	S1.cnt = 0;
	S1.encnt = 0;
}

void GAMEBOY::s2_init()
{
	S2.envol = R_NR22 >> 4;
	S2.endir = (R_NR22>>3) & 1;
	S2.endir |= S2.endir - 1;
	S2.enlen = (R_NR22 & 7) << 15;
	if (!S2.on) S2.pos = 0;
	S2.on = 1;
	S2.cnt = 0;
	S2.encnt = 0;
}

void GAMEBOY::s3_init()
{
	int i;
	if (!S3.on) S3.pos = 0;
	S3.cnt = 0;
	S3.on = R_NR30 >> 7;
	if (S3.on) for (i = 0; i < 16; i++)
		Hi_Memory[i+0x130] = 0x13 ^ Hi_Memory[i+0x131];
}

void GAMEBOY::s4_init()
{
	S4.envol = R_NR42 >> 4;
	S4.endir = (R_NR42>>3) & 1;
	S4.endir |= S4.endir - 1;
	S4.enlen = (R_NR42 & 7) << 15;
	S4.on = 1;
	S4.pos = 0;
	S4.cnt = 0;
	S4.encnt = 0;
}


void GAMEBOY::sound_write(int r, byte b)
{
#if 0
	static void *timer;
	if (!timer) timer = sys_timer();
	printf("write %02X: %02X @ %d\n", r, b, sys_elapsed(timer));
#endif

#ifdef GB_SOUND_GNUBOY
	if (!(R_NR52 & 128) && r != RI_NR52) return;
	if ((r & 0xF0) == 0x30)
	{
		if (S3.on) sound_mix();
		if (!S3.on)
			WAVE[(r&0xFF)-0x30] = Hi_Memory[0x100+(r&0xFF)] = b;
		return;
	}
//	sound_mix();
	switch (r)
	{
	case RI_NR10:
		R_NR10 = b;
		S1.swlen = ((R_NR10>>4) & 7) << 14;
		S1.swfreq = ((R_NR14&7)<<8) + R_NR13;
		break;
	case RI_NR11:
		R_NR11 = b;
		S1.len = (64-(R_NR11&63)) << 13;
		break;
	case RI_NR12:
		R_NR12 = b;
		S1.envol = R_NR12 >> 4;
		S1.endir = (R_NR12>>3) & 1;
		S1.endir |= S1.endir - 1;
		S1.enlen = (R_NR12 & 7) << 15;
		break;
	case RI_NR13:
		R_NR13 = b;
		s1_freq();
		break;
	case RI_NR14:
		R_NR14 = b;
		s1_freq();
		if (b & 128) s1_init();
		break;
	case RI_NR21:
		R_NR21 = b;
		S2.len = (64-(R_NR21&63)) << 13;
		break;
	case RI_NR22:
		R_NR22 = b;
		S2.envol = R_NR22 >> 4;
		S2.endir = (R_NR22>>3) & 1;
		S2.endir |= S2.endir - 1;
		S2.enlen = (R_NR22 & 7) << 15;
		break;
	case RI_NR23:
		R_NR23 = b;
		s2_freq();
		break;
	case RI_NR24:
		R_NR24 = b;
		s2_freq();
		if (b & 128) s2_init();
		break;
	case RI_NR30:
		R_NR30 = b;
		if (!(b & 128)) S3.on = 0;
		break;
	case RI_NR31:
		R_NR31 = b;
		S3.len = (256-R_NR31) << 13;
		break;
	case RI_NR32:
		R_NR32 = b;
		break;
	case RI_NR33:
		R_NR33 = b;
		s3_freq();
		break;
	case RI_NR34:
		R_NR34 = b;
		s3_freq();
		if (b & 128) s3_init();
		break;
	case RI_NR41:
		R_NR41 = b;
		S4.len = (64-(R_NR41&63)) << 13;
		break;
	case RI_NR42:
		R_NR42 = b;
		S4.envol = R_NR42 >> 4;
		S4.endir = (R_NR42>>3) & 1;
		S4.endir |= S4.endir - 1;
		S4.enlen = (R_NR42 & 7) << 15;
		break;
	case RI_NR43:
		R_NR43 = b;
		s4_freq();
		break;
	case RI_NR44:
		R_NR44 = b;
		if (b & 128) s4_init();
		break;
	case RI_NR50:
		R_NR50 = b;
		break;
	case RI_NR51:
		R_NR51 = b;
		break;
	case RI_NR52:
		R_NR52 = b;
		if (!(R_NR52 & 128))
			sound_off();
		break;
	default:
		return;
	}
#else
	//sndRegChange = 1;
	Hi_Memory[r]=b;
#endif
}



void GAMEBOY::apu_enqueue(gb_apudata_t *d)
{
	/*
	if(apu_qwritep==apu_qreadp){
#ifdef _DEBUG
		LOG("apu_enqueue error" <<endl);
#endif
//		return;
	}
	*/
	apu_queue[apu_qwritep] = *d;
	apu_qwritep= (apu_qwritep+1)&GB_APUQUEUE_MASK;
}


gb_apudata_t *GAMEBOY::apu_dequeue(void)
{
#ifdef _DEBUG
//	LOG("apu_qreadp " << HEX(apu_qreadp,8) <<"  apu_qwritep " << HEX(apu_qwritep,8) <<endl);
#endif
//	if(apu_qreadp == apu_qwritep)
//		return NULL;
	gb_apudata_t *t;
	t = &apu_queue[apu_qreadp];
	apu_qreadp= (apu_qreadp+1)&GB_APUQUEUE_MASK;
	return t;
}


static int bit[4][8]={{1,0,0,0,0,0,0},{1,1,0,0,0,0,0,0},{1,1,1,1,0,0,0,0},{1,1,1,1,1,1,0,0}};
static int vol[16]={0,8,17,25,34,42,51,59,68,76,85,93,102,110,119,127};

#define HZ(x) ((8192*1024)/(x))
//static int bufPos,bufCycles,bufLVal,bufRVal;
//static long sampleCycles= (8192*1024)/SRATE; //(8192*1024)/SRATE; //(8192*1024)/SRATE; //22050; //22050;




void GAMEBOY::apu_process(){
	unsigned char *buffer;
	int num_samples;
	int elapsed_cycles, inc_cycles, buffpos=0;
	int s, l, r, f, n, i;
	static int prelv=0, prerv=0;
	gb_apudata_t *d;

	elapsed_cycles = 0;

    if(snd_mgr->IsNull())
		return;
    {
		sound_buf_pos cur_ph;  // cur playing half
		sound_buf_pos cur_nph; // cur not-playing half
		uint32 buf_len;

		cur_ph = snd_mgr->get_currently_playing_half();
		if(cur_ph == currently_playing_half) return;
		cur_nph = currently_playing_half;
		currently_playing_half = cur_ph;
		if(!snd_mgr->lock(cur_nph, (void**)&buffer, &buf_len)){
			apu_qreadp = apu_qwritep =0;
			gb_cpu.snd = 0;
			pcm.pos = 0;
//			LOG("couldn't lock sound buffer" << endl);
			return;
		}
#ifdef _DEBUG
//		LOG("lock sound buffer " << HEX(buf_len,8) <<endl);
//		gb_sndclock = 0;
#endif
		num_samples = buf_len/(snd_mgr->get_sample_bits()/8*2);
//		memcpy(buf, pcm.buf, buf_len);
    }
//	for (; gb_cpu.snd >= RATE; gb_cpu.snd -= RATE)
	inc_cycles = snd.rate; //4194304 / num_samples;//<<1;
	for(i=0;i<num_samples;i++, elapsed_cycles+=inc_cycles){
		l = r = 0;

		while((apu_qreadp != apu_qwritep) && (apu_queue[apu_qreadp].timestamp <= elapsed_cycles)){
			d = apu_dequeue();
			sound_write(d->address, d->value);
		}
#ifdef GB_SOUND_GNUBOY
		if (S1.on)
		{
			s = sqwave[R_NR11>>6][(S1.pos>>18)&7] & S1.envol;
			S1.pos += S1.freq;
			if ((R_NR14 & 64) && ((S1.cnt += RATE) >= S1.len))
				S1.on = 0;
			if (S1.enlen && (S1.encnt += RATE) >= S1.enlen)
			{
				S1.encnt -= S1.enlen;
				S1.envol += S1.endir;
				if (S1.envol < 0) S1.envol = 0;
				if (S1.envol > 15) S1.envol = 15;
			}
			if (S1.swlen && (S1.swcnt += RATE) >= S1.swlen)
			{
				S1.swcnt -= S1.swlen;
				f = S1.swfreq;
				n = (R_NR10 & 7);
				if (R_NR10 & 8) f -= (f >> n);
				else f += (f >> n);
				if (f > 2047)
					S1.on = 0;
				else
				{
					S1.swfreq = f;
					R_NR13 = f;
					R_NR14 = (R_NR14 & 0xF8) | (f>>8);
					s1_freq_d(2048 - f);
				}
			}
			s <<= 2;
			if (R_NR51 & 1) r += s;
			if (R_NR51 & 16) l += s;
		}
		
		if (S2.on)
		{
			s = sqwave[R_NR21>>6][(S2.pos>>18)&7] & S2.envol;
			S2.pos += S2.freq;
			if ((R_NR24 & 64) && ((S2.cnt += RATE) >= S2.len))
				S2.on = 0;
			if (S2.enlen && (S2.encnt += RATE) >= S2.enlen)
			{
				S2.encnt -= S2.enlen;
				S2.envol += S2.endir;
				if (S2.envol < 0) S2.envol = 0;
				if (S2.envol > 15) S2.envol = 15;
			}
			s <<= 2;
			if (R_NR51 & 2) r += s;
			if (R_NR51 & 32) l += s;
		}
		
		if (S3.on)
		{
			s = WAVE[(S3.pos>>22) & 15];
			if (S3.pos & (1<<21)) s &= 15;
			else s >>= 4;
			s -= 8;
			S3.pos += S3.freq;
			if ((R_NR34 & 64) && ((S3.cnt += RATE) >= S3.len))
				S3.on = 0;
			if (R_NR32 & 96) s <<= (3 - ((R_NR32>>5)&3));
			else s = 0;
			if (R_NR51 & 4) r += s;
			if (R_NR51 & 64) l += s;
		}
		
		if (S4.on)
		{
			if (R_NR43 & 8) s = 1 & (noise7[(S4.pos>>20)&15] >> (7-((S4.pos>>17)&7)));
			else s = 1 & (noise15[(S4.pos>>20)&4095] >> (7-((S4.pos>>17)&7)));
			s = (-s) & S4.envol;
			S4.pos += S4.freq;
			if ((R_NR44 & 64) && ((S4.cnt += RATE) >= S4.len))
				S4.on = 0;
			if (S4.enlen && (S4.encnt += RATE) >= S4.enlen)
			{
				S4.encnt -= S4.enlen;
				S4.envol += S4.endir;
				if (S4.envol < 0) S4.envol = 0;
				if (S4.envol > 15) S4.envol = 15;
			}
			s += s << 1;
			if (R_NR51 & 8) r += s;
			if (R_NR51 & 128) l += s;
		}
		
		l *= (R_NR50 & 0x07);
		r *= ((R_NR50 & 0x70)>>4);
		l >>= 4;
		r >>= 4;
		
		if (l > 127) l = 127;
		else if (l < -128) l = -128;
		if (r > 127) r = 127;
		else if (r < -128) r = -128;
		
		if(buffer){
			if(pcm.stereo){
				// Low pass filter
				int tmpr, tmpl;
/*
				{
					tmpr = r; tmpl = l;
					l += prelv;
					l >>=1;
					r += prerv;
					r >>=1;
				}
*/
				buffer[buffpos++] = l+128;
				buffer[buffpos++] = r+128;
				prelv = tmpl;
				prerv = tmpr;
			}
			else{
				l += prelv;
				l >>=1;
				buffer[buffpos++] = ((l+r)>>1)+128;
			}
		}
		R_NR52 = (R_NR52&0xf0) | S1.on | (S2.on<<1) | (S3.on<<2) | (S4.on<<3);
#else
		int cycles = inc_cycles<<2;

		if(!(REG_FF26&0x80)){
			buffer[buffpos++] = 128;
			buffer[buffpos++] = 128;
		}
//		if(sndRegChange)
		else
		{
/*
			snd1Enable = 1;
			snd2Enable = 1;
			snd3Enable = 1;
			snd4Enable = 1;
*/

			snd1Freq=REG_FF13|((REG_FF14&7)<<8);
			snd2Freq=REG_FF18|((REG_FF19&7)<<8);
			snd3Freq=REG_FF1D|((REG_FF1E&7)<<8);
			switch(REG_FF22&7)
			{
            case 0: snd4Freq=((512*1024)*2)>>((REG_FF22>>4)+1); break;
            case 1: snd4Freq=(512*1024)>>((REG_FF22>>4)+1); break;
            case 2: snd4Freq=((512*1024)/2)>>((REG_FF22>>4)+1); break;
            case 3: snd4Freq=((512*1024)/3)>>((REG_FF22>>4)+1); break;
            case 4: snd4Freq=((512*1024)/4)>>((REG_FF22>>4)+1); break;
            case 5: snd4Freq=((512*1024)/5)>>((REG_FF22>>4)+1); break;
            case 6: snd4Freq=((512*1024)/6)>>((REG_FF22>>4)+1); break;
            case 7: snd4Freq=((512*1024)/7)>>((REG_FF22>>4)+1); break;
			}
			snd4Freq=HZ(snd4Freq);
			snd3Enable=(REG_FF1A&0x80)?1:0;
			snd1Vol=REG_FF12>>4;
			if(!(REG_FF26&1)){
            	snd1Enable=0;
			}
			else
			if (REG_FF14&0x80)
			{
				snd1Len=64-(REG_FF11&63);
				snd1Cnt=0;
				REG_FF14&=0x7f;
				snd1Enable=1;
			}
			snd2Vol=REG_FF17>>4;
			if(!(REG_FF26&2)){
            	snd2Enable=0;
			}
			else
			if (REG_FF19&0x80)
			{
				snd2Len=64-(REG_FF16&63);
				snd2Cnt=0;
				REG_FF19&=0x7f;
				snd2Enable=1;
			}
			if (REG_FF1E&0x80)
			{
				snd3Len=(256-((int)((unsigned char)REG_FF1B)))<<7;
				snd3Cnt=0;
				REG_FF1E&=0x7f;
			}
			snd4Vol=REG_FF21>>4;
			if(!(REG_FF26&8)){
				snd4Enable=0;
			}
			else
			if (REG_FF23&0x80)
			{
				snd4Len=64-(REG_FF20&63);
				REG_FF23&=0x7f;
				snd4Enable=1;
			}
//			sndRegChange=0;
		}
		if ((snd1Enable)&&(REG_FF10&0x70))
		{
			swpClk+=cycles;
			if (swpClk>=HZ(128))
			{
				swpClk-=HZ(128);
				snd1SwpCnt++;
				if (snd1SwpCnt>=((REG_FF10>>4)&7))
				{
					snd1SwpCnt=0;
					if (REG_FF10&8)
					{
						snd1Freq-=snd1Freq>>(REG_FF10&7);
						if (snd1Freq<0) snd1Freq=0;
					}
					else
					{
						snd1Freq+=snd1Freq>>(REG_FF10&7);
						if (snd1Freq>2047)
						{
							snd1Freq=2047;
							snd1Enable=0;
						}
					}
				}
			}
		}
		envClk+=cycles;
		if (envClk>=HZ(64))
		{
			envClk-=HZ(64);
			if ((snd1Enable)&&(REG_FF12&7))
			{
				snd1EnvCnt++;
				if (snd1EnvCnt>=(REG_FF12&7))
				{
					snd1EnvCnt=0;
					if (REG_FF12&8)
					{
						snd1Vol++;
						if (snd1Vol>0xf) snd1Vol=0xf;
						REG_FF12=(REG_FF12&0xf)|(snd1Vol<<4);
					}
					else
					{
						snd1Vol--;
						if (snd1Vol<0) snd1Vol=0;
						REG_FF12=(REG_FF12&0xf)|(snd1Vol<<4);
					}
				}
			}
			if ((snd2Enable)&&(REG_FF17&7))
			{
				snd2EnvCnt++;
				if (snd2EnvCnt>=(REG_FF17&7))
				{
					snd2EnvCnt=0;
					if (REG_FF17&8)
					{
						snd2Vol++;
						if (snd2Vol>0xf) snd2Vol=0xf;
						REG_FF17=(REG_FF17&0xf)|(snd2Vol<<4);
					}
					else
					{
						snd2Vol--;
						if (snd2Vol<0) snd2Vol=0;
						REG_FF17=(REG_FF17&0xf)|(snd2Vol<<4);
					}
				}
			}
			if ((snd4Enable)&&(REG_FF21&7))
			{
				snd4EnvCnt++;
				if (snd4EnvCnt>=(REG_FF21&7))
				{
					snd4EnvCnt=0;
					if (REG_FF21&8)
					{
						snd4Vol++;
						if (snd4Vol>0xf) snd4Vol=0xf;
						REG_FF21=(REG_FF21&0xf)|(snd4Vol<<4);
					}
					else
					{
						snd4Vol--;
						if (snd4Vol<0) snd4Vol=0;
						REG_FF21=(REG_FF21&0xf)|(snd4Vol<<4);
					}
				}
			}
		}
		lenClk+=cycles;
		if (lenClk>=HZ(256))
		{
			lenClk-=HZ(256);
			if (snd1Enable)
			{
				snd1Len--;
				if ((snd1Len<=0)&&(REG_FF14&0x40))
					snd1Enable=0;
			}
			if (snd2Enable)
			{
				snd2Len--;
				if ((snd2Len<=0)&&(REG_FF19&0x40))
					snd2Enable=0;
			}
			if (snd3Enable)
			{
				snd3Len--;
				if ((snd3Len<=0)&&(REG_FF1E&0x40))
				{
					snd3Enable=0;
					REG_FF1A&=0x7f;
				}
			}
			if (snd4Enable)
			{
				snd4Len--;
				if ((snd4Len<=0)&&(REG_FF23&0x40))
					snd4Enable=0;
			}
		}
		REG_FF13=snd1Freq&0xff;
		REG_FF14=(REG_FF14&0xf8)|((snd1Freq>>8)&7);
		freqClk+=cycles;
		if (freqClk>=4)
		{
			int n=freqClk>>2;
			freqClk-=n<<2;
			if (snd1Enable)
			{
				snd1Cnt+=n;
				while (snd1Cnt>=((2048-snd1Freq)<<4))
					snd1Cnt-=((2048-snd1Freq)<<4);
			}
			if (snd2Enable)
			{
				snd2Cnt+=n;
				while (snd2Cnt>=((2048-snd2Freq)<<4))
					snd2Cnt-=((2048-snd2Freq)<<4);
			}
			if (snd3Enable)
			{
				snd3Cnt+=n;
				while (snd3Cnt>=((2048-snd3Freq)<<5))
					snd3Cnt-=((2048-snd3Freq)<<5);
			}
		}
		if (snd1Enable)
		{
			int stage=(snd1Cnt/(2048-snd1Freq))>>1;
			if (stage>7) stage=7;
			snd1Bit=bit[REG_FF11>>6][stage];
		}
		if (REG_FF25&1) l+=snd1Bit?vol[snd1Vol]:-vol[snd1Vol];
		if (REG_FF25&0x10) r+=snd1Bit?vol[snd1Vol]:-vol[snd1Vol];
		if (snd2Enable)
		{
			int stage=(snd2Cnt/(2048-snd2Freq))>>1;
			if (stage>7) stage=7;
			snd2Bit=bit[REG_FF16>>6][stage];
		}
		if (REG_FF25&2) l+=snd2Bit?vol[snd2Vol]:-vol[snd2Vol];
		if (REG_FF25&0x20) r+=snd2Bit?vol[snd2Vol]:-vol[snd2Vol];
		if (snd3Enable)
		{
			int stage=snd3Cnt/(2048-snd3Freq);
			if (stage>31) stage=31;
			snd3Bit=Hi_Memory[0x130+(stage>>1)];	//FF30
			if (stage&1)
				snd3Bit&=0xf;
			else
				snd3Bit>>=4;
			switch ((REG_FF1C>>5)&3)
			{
            case 0: snd3Bit=8; break;
            case 1: break;
            case 2: snd3Bit=8|(snd3Bit>>1); break;
            case 3: snd3Bit=0xc|(snd3Bit>>2); break;
			}
		}
		if (REG_FF25&4) l+=(snd3Bit<<4)-0x80;
		if (REG_FF25&0x40) r+=(snd3Bit<<4)-0x80;
		if (snd4Enable)
		{
			freq4Clk+=cycles;
			if (freq4Clk>=snd4Freq)
			{
				freq4Clk=freq4Clk%snd4Freq;
				snd4Bit=rand()&1;
			}
		}
		if (REG_FF25&8) l+=snd4Bit?vol[snd4Vol]:-vol[snd4Vol];
		if (REG_FF25&0x80) r+=snd4Bit?vol[snd4Vol]:-vol[snd4Vol];
		l>>=2;
		if (l<-128) l=-128;
		if (l>127) l=127;
//		l+=0x80;
		r>>=2;
		if (r<-128) r=-128;
		if (r>127) r=127;
//		r+=0x80;
//#if 1
		{
			// Low pass filter
			int tmpr, tmpl;
			tmpr = r; tmpl = l;
			l += prelv;
			l >>=1;
			r += prerv;
			r >>=1;
			buffer[buffpos++] = l+128;
			buffer[buffpos++] = r+128;
			prelv = tmpl;
			prerv = tmpr;
		}
#endif
	}
	snd_mgr->unlock();
#ifdef _DEBUG
//	LOG("__ apu_qreadp " << HEX(apu_qreadp,8) <<"  apu_qwritep " << HEX(apu_qwritep,8) <<endl);
#endif
	{
		int tmp;
		for(tmp = apu_qreadp; tmp != apu_qwritep; tmp= (tmp+1)&GB_APUQUEUE_MASK){
			apu_queue[tmp].timestamp -= elapsed_cycles/*-inc_cycles*/;
		}
	}
//	apu_qreadp = apu_qwritep =0;
	gb_cpu.snd = 0;
	pcm.pos = 0;
}


void GAMEBOY::GetROMInfoStr(char *wt){
	const char *mappertype[] = {"MBC1","MBC2","MBC3","RUMBLE","MBC5","HUC1","HUC3", "na"};
	int ts;

	switch (mbc_type)
	{
	case MBC_MBC1: ts = 0; break;
	case MBC_MBC2: ts = 1; break;
	case MBC_MBC3: ts = 2; break;
	case MBC_RUMBLE: ts = 3; break;
	case MBC_MBC5: ts = 4; break;
	case MBC_HUC1: ts = 5; break;
	case MBC_HUC3: ts = 6; break;
	default: ts = 7; break;
	}
	wsprintf(wt, "%s", mappertype[ts]);
}


void GAMEBOY::GetCPUContext(unsigned char *context){
	NES_6502::Context *nes_context = (NES_6502::Context *)context;

	nes_context->mem_page[0] = ROMimage;
	nes_context->mem_page[1] = ROMimage + 0x2000;
	nes_context->mem_page[2] = gb_cpu.mem_page[0x04];
	nes_context->mem_page[3] = gb_cpu.mem_page[0x04]+0x2000;
	nes_context->mem_page[5] = gb_cpu.mem_page[0x0a];
	nes_context->mem_page[6] = MainMem;
}


#endif
